from compressor.conf import settings
from compressor.filters import CompilerFilter


class YUICompressorFilter(CompilerFilter):
    command = "{binary} {args}"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.command += " --type=%s" % self.type
        if self.verbose:
            self.command += " --verbose"


class YUICSSFilter(YUICompressorFilter):
    type = "css"
    options = (
        ("binary", settings.COMPRESS_YUI_BINARY),
        ("args", settings.COMPRESS_YUI_CSS_ARGUMENTS),
    )


class YUIJSFilter(YUICompressorFilter):
    type = "js"
    options = (
        ("binary", settings.COMPRESS_YUI_BINARY),
        ("args", settings.COMPRESS_YUI_JS_ARGUMENTS),
    )



from django.test import SimpleTestCase
from django.test.utils import override_settings
from compressor.conf import settings
from compressor.conf import CompressorConf


default_css_filters = [
    "compressor.filters.css_default.CssAbsoluteFilter",
    "compressor.filters.cssmin.rCSSMinFilter",
]
default_js_filters = ["compressor.filters.jsmin.rJSMinFilter"]


def create_conf(**attrs):
    # Creating a new appconf.AppConf subclass will cause
    # its configuration to be resolved.
    # We use this to force the CompressorConf to be re-resolved,
    # when we've changed the settings.
    attrs["__module__"] = None
    return type("TestCompressorConf", (CompressorConf,), attrs)


class ConfTestCase(SimpleTestCase):
    def test_filter_defaults(self):
        # This used the settings from compressor/test_settings.py
        # which contains no values for filers and therefore uses the defaults.
        self.assertEqual(settings.COMPRESS_FILTERS["css"], default_css_filters)
        self.assertEqual(settings.COMPRESS_FILTERS["js"], default_js_filters)

    @override_settings(COMPRESS_FILTERS=dict(css=["ham"], js=["spam"]))
    def test_filters_by_main_setting(self):
        conf = create_conf()
        self.assertEqual(conf.FILTERS["css"], ["ham"])
        self.assertEqual(conf.FILTERS["js"], ["spam"])



from compressor.utils import staticfiles
from compressor.storage import CompressorFileStorage


class CompressorFinder(staticfiles.finders.BaseStorageFinder):
    """
    A staticfiles finder that looks in COMPRESS_ROOT
    for compressed files, to be used during development
    with staticfiles development file server or during
    deployment.
    """

    storage = CompressorFileStorage

    def list(self, ignore_patterns):
        return []



# flake8: noqa
import os
import sys
import concurrent.futures
from threading import Lock

from collections import OrderedDict, defaultdict
from fnmatch import fnmatch
from importlib import import_module

import django
from django.core.management.base import BaseCommand, CommandError
import django.template
from django.template import Context
from django.utils.encoding import smart_str
from django.template.loader import (
    get_template,
)  # noqa Leave this in to preload template locations
from django.template import engines

from compressor.cache import (
    get_offline_hexdigest,
    write_offline_manifest,
    get_offline_manifest,
)
from compressor.conf import settings
from compressor.exceptions import (
    OfflineGenerationError,
    TemplateSyntaxError,
    TemplateDoesNotExist,
)
from compressor.utils import get_mod_func

offline_manifest_lock = Lock()
node_locks_lock = Lock()
node_locks = defaultdict(Lock)


class Command(BaseCommand):
    help = "Compress content outside of the request/response cycle"

    requires_system_checks = []

    def add_arguments(self, parser):
        parser.add_argument(
            "--extension",
            "-e",
            action="append",
            dest="extensions",
            help='The file extension(s) to examine (default: ".html", '
            "separate multiple extensions with commas, or use -e "
            "multiple times)",
        )
        parser.add_argument(
            "-f",
            "--force",
            default=False,
            action="store_true",
            help="Force the generation of compressed content even if the "
            "COMPRESS_ENABLED setting is not True.",
            dest="force",
        )
        parser.add_argument(
            "--follow-links",
            default=False,
            action="store_true",
            help="Follow symlinks when traversing the COMPRESS_ROOT "
            "(which defaults to STATIC_ROOT). Be aware that using this "
            "can lead to infinite recursion if a link points to a parent "
            "directory of itself.",
            dest="follow_links",
        )
        parser.add_argument(
            "--engine",
            default=[],
            action="append",
            help="Specifies the templating engine. jinja2 and django are "
            "supported. It may be a specified more than once for "
            "multiple engines. If not specified, django engine is used.",
            dest="engines",
        )

    def get_loaders(self):
        template_source_loaders = []
        for e in engines.all():
            if hasattr(e, "engine"):
                template_source_loaders.extend(
                    e.engine.get_template_loaders(e.engine.loaders)
                )
        loaders = []
        # If template loader is CachedTemplateLoader, return the loaders
        # that it wraps around. So if we have
        # TEMPLATE_LOADERS = (
        #    ('django.template.loaders.cached.Loader', (
        #        'django.template.loaders.filesystem.Loader',
        #        'django.template.loaders.app_directories.Loader',
        #    )),
        # )
        # The loaders will return django.template.loaders.filesystem.Loader
        # and django.template.loaders.app_directories.Loader
        # The cached Loader and similar ones include a 'loaders' attribute
        # so we look for that.
        for loader in template_source_loaders:
            if hasattr(loader, "loaders"):
                loaders.extend(loader.loaders)
            else:
                loaders.append(loader)
        return loaders

    def __get_parser(self, engine):
        charset = (
            settings.FILE_CHARSET if settings.is_overridden("FILE_CHARSET") else "utf-8"
        )
        if engine == "jinja2":
            from compressor.offline.jinja2 import Jinja2Parser

            env = settings.COMPRESS_JINJA2_GET_ENVIRONMENT()
            parser = Jinja2Parser(charset=charset, env=env)
        elif engine == "django":
            from compressor.offline.django import DjangoParser

            parser = DjangoParser(charset=charset)
        else:
            raise OfflineGenerationError("Invalid templating engine specified.")

        return parser

    def compress(self, engine, extensions, verbosity, follow_links, log):
        """
        Searches templates containing 'compress' nodes and compresses them
        "offline" -- outside of the request/response cycle.

        The result is cached with a cache-key derived from the content of the
        compress nodes (not the content of the possibly linked files!).
        """

        if not self.get_loaders():
            raise OfflineGenerationError(
                "No template loaders defined. You "
                "must set TEMPLATE_LOADERS in your "
                "settings or set 'loaders' in your "
                "TEMPLATES dictionary."
            )
        templates = set()
        if engine == "django":
            paths = set()
            for loader in self.get_loaders():
                try:
                    module = import_module(loader.__module__)
                    get_template_sources = getattr(module, "get_template_sources", None)
                    if get_template_sources is None:
                        get_template_sources = loader.get_template_sources
                    paths.update(
                        smart_str(origin) for origin in get_template_sources("")
                    )
                except (ImportError, AttributeError, TypeError):
                    # Yeah, this didn't work out so well, let's move on
                    pass

            if not paths:
                raise OfflineGenerationError(
                    "No template paths found. None of "
                    "the configured template loaders "
                    "provided template paths. See "
                    "https://docs.djangoproject.com/en/2.1/topics/templates/ "
                    "for more information on template "
                    "loaders."
                )
            if verbosity >= 2:
                log.write("Considering paths:\n\t" + "\n\t".join(paths) + "\n")

            for path in paths:
                for root, dirs, files in os.walk(path, followlinks=follow_links):
                    templates.update(
                        os.path.relpath(os.path.join(root, name), path)
                        for name in files
                        if not name.startswith(".")
                        and any(fnmatch(name, "*%s" % glob) for glob in extensions)
                    )
        elif engine == "jinja2":
            env = settings.COMPRESS_JINJA2_GET_ENVIRONMENT()
            if env and hasattr(env, "list_templates"):
                templates |= set(
                    [
                        env.loader.get_source(env, template)[1]
                        for template in env.list_templates(
                            filter_func=lambda _path: os.path.splitext(_path)[-1]
                            in extensions
                        )
                    ]
                )

        if not templates:
            raise OfflineGenerationError(
                "No templates found. Make sure your "
                "TEMPLATE_LOADERS and TEMPLATE_DIRS "
                "settings are correct."
            )
        if verbosity >= 2:
            log.write("Found templates:\n\t" + "\n\t".join(templates) + "\n")

        contexts = settings.COMPRESS_OFFLINE_CONTEXT
        if isinstance(contexts, str):
            try:
                module, function = get_mod_func(contexts)
                contexts = getattr(import_module(module), function)()
            except (AttributeError, ImportError, TypeError) as e:
                raise ImportError(
                    "Couldn't import offline context function %s: %s"
                    % (settings.COMPRESS_OFFLINE_CONTEXT, e)
                )
        elif not isinstance(contexts, (list, tuple)):
            contexts = [contexts]

        parser = self.__get_parser(engine)
        fine_templates = []

        if verbosity >= 1:
            log.write("Compressing... ")

        for template_name in templates:
            try:
                template = parser.parse(template_name)
                template.template_name = template_name
                fine_templates.append(template)
            except IOError:  # unreadable file -> ignore
                if verbosity >= 1:
                    log.write("Unreadable template at: %s\n" % template_name)
                continue
            except TemplateSyntaxError as e:  # broken template -> ignore
                if verbosity >= 1:
                    log.write(
                        "Invalid template %s: %s\n" % (template_name, smart_str(e))
                    )
                continue
            except TemplateDoesNotExist:  # non existent template -> ignore
                if verbosity >= 1:
                    log.write("Non-existent template at: %s\n" % template_name)
                continue
            except UnicodeDecodeError:
                if verbosity >= 1:
                    log.write(
                        "UnicodeDecodeError while trying to read "
                        "template %s\n" % template_name
                    )
                continue

        contexts_count = 0
        nodes_count = 0
        offline_manifest = OrderedDict()
        errors = []

        for context_dict in contexts:
            compressor_nodes = OrderedDict()
            for template in fine_templates:
                context = Context(parser.get_init_context(context_dict))

                try:
                    nodes = list(parser.walk_nodes(template, context=context))
                except (TemplateDoesNotExist, TemplateSyntaxError) as e:
                    # Could be an error in some base template
                    if verbosity >= 1:
                        log.write(
                            "Error parsing template %s: %s\n"
                            % (template.template_name, smart_str(e))
                        )
                    continue

                if nodes:
                    template_nodes = compressor_nodes.setdefault(
                        template, OrderedDict()
                    )
                    for node in nodes:
                        nodes_count += 1
                        template_nodes.setdefault(node, []).append(context)

            pool = concurrent.futures.ThreadPoolExecutor(max_workers=4)
            for template, nodes in compressor_nodes.items():
                template._log = log
                template._log_verbosity = verbosity

                pool.submit(
                    self._compress_template,
                    offline_manifest,
                    nodes,
                    parser,
                    template,
                    errors,
                )

            pool.shutdown(wait=True)
            contexts_count += 1

        # If errors exist, raise the first one in the list
        if errors:
            raise errors[0]
        elif not nodes_count:
            raise OfflineGenerationError(
                "No 'compress' template tags found in templates."
                "Try running compress command with --follow-links and/or"
                "--extension=EXTENSIONS"
            )

        if verbosity >= 1:
            log.write(
                "done\nCompressed %d block(s) from %d template(s) for %d context(s).\n"
                % (len(offline_manifest), nodes_count, contexts_count)
            )
        return offline_manifest, len(offline_manifest), offline_manifest.values()

    @staticmethod
    def _compress_template(offline_manifest, nodes, parser, template, errors):
        for node, node_contexts in nodes.items():
            for context in node_contexts:
                context.push()
                if not parser.process_template(template, context):
                    continue

                parser.process_node(template, context, node)
                rendered = parser.render_nodelist(template, context, node)
                key = get_offline_hexdigest(rendered)

                # Atomically check if the key exists in offline manifest.
                # If it doesn't, set a placeholder key (None). This is to prevent
                # concurrent _compress_template instances from rendering the
                # same node, and then writing to the same file.
                with offline_manifest_lock:
                    if key in offline_manifest:
                        continue

                    offline_manifest[key] = None

                node_id = id(node)

                with node_locks_lock:
                    node_lock = node_locks[node_id]

                try:
                    with node_lock:
                        result = parser.render_node(template, context, node)
                except Exception as e:
                    errors.append(
                        CommandError(
                            "An error occurred during rendering %s: "
                            "%s" % (template.template_name, smart_str(e))
                        )
                    )
                    del offline_manifest[key]
                    return
                finally:
                    with node_locks_lock:
                        node_locks.pop(node_id, None)
                result = result.replace(
                    settings.COMPRESS_URL, settings.COMPRESS_URL_PLACEHOLDER
                )
                offline_manifest[key] = result
                context.pop()

    def handle_extensions(self, extensions=("html",)):
        """
        organizes multiple extensions that are separated with commas or
        passed by using --extension/-e multiple times.

        for example: running 'django-admin compress -e js,txt -e xhtml -a'
        would result in an extension list: ['.js', '.txt', '.xhtml']

        >>> handle_extensions(['.html', 'html,js,py,py,py,.py', 'py,.py'])
        ['.html', '.js']
        >>> handle_extensions(['.html, txt,.tpl'])
        ['.html', '.tpl', '.txt']
        """
        ext_list = []
        for ext in extensions:
            ext_list.extend(ext.replace(" ", "").split(","))
        for i, ext in enumerate(ext_list):
            if not ext.startswith("."):
                ext_list[i] = ".%s" % ext_list[i]
        return set(ext_list)

    def handle(self, **options):
        self.handle_inner(**options)

    def handle_inner(self, **options):
        if not settings.COMPRESS_ENABLED and not options.get("force"):
            raise CommandError(
                "Compressor is disabled. Set the COMPRESS_ENABLED "
                "setting or use --force to override."
            )
        if not settings.COMPRESS_OFFLINE:
            if not options.get("force"):
                raise CommandError(
                    "Offline compression is disabled. Set "
                    "COMPRESS_OFFLINE or use the --force to override."
                )

        log = options.get("log", sys.stdout)
        verbosity = options.get("verbosity", 1)
        follow_links = options.get("follow_links", False)
        extensions = self.handle_extensions(options.get("extensions") or ["html"])
        engines = [e.strip() for e in options.get("engines", [])] or ["django"]

        final_offline_manifest = {}
        final_block_count = 0
        final_results = []
        for engine in engines:
            offline_manifest, block_count, results = self.compress(
                engine, extensions, verbosity, follow_links, log
            )
            final_results.extend(results)
            final_block_count += block_count
            final_offline_manifest.update(offline_manifest)
        write_offline_manifest(final_offline_manifest)
        return final_block_count, final_results



import fnmatch
import os

from django.core.management.base import BaseCommand, CommandError

from compressor.conf import settings
from compressor.cache import cache, get_mtime, get_mtime_cachekey


class Command(BaseCommand):
    help = "Add or remove all mtime values from the cache"

    def add_arguments(self, parser):
        parser.add_argument(
            "-i",
            "--ignore",
            action="append",
            default=[],
            dest="ignore_patterns",
            metavar="PATTERN",
            help="Ignore files or directories matching this glob-style "
            "pattern. Use multiple times to ignore more.",
        ),
        parser.add_argument(
            "--no-default-ignore",
            action="store_false",
            dest="use_default_ignore_patterns",
            default=True,
            help="Don't ignore the common private glob-style patterns 'CVS', "
            "'.*' and '*~'.",
        ),
        parser.add_argument(
            "--follow-links",
            dest="follow_links",
            action="store_true",
            help="Follow symlinks when traversing the COMPRESS_ROOT "
            "(which defaults to STATIC_ROOT). Be aware that using this "
            "can lead to infinite recursion if a link points to a parent "
            "directory of itself.",
        ),
        parser.add_argument(
            "-c", "--clean", dest="clean", action="store_true", help="Remove all items"
        ),
        parser.add_argument(
            "-a", "--add", dest="add", action="store_true", help="Add all items"
        ),

    def is_ignored(self, path):
        """
        Return True or False depending on whether the ``path`` should be
        ignored (if it matches any pattern in ``ignore_patterns``).
        """
        for pattern in self.ignore_patterns:
            if fnmatch.fnmatchcase(path, pattern):
                return True
        return False

    def handle(self, **options):
        ignore_patterns = options["ignore_patterns"]
        if options["use_default_ignore_patterns"]:
            ignore_patterns += ["CVS", ".*", "*~"]
            options["ignore_patterns"] = ignore_patterns
        self.ignore_patterns = ignore_patterns

        if (options["add"] and options["clean"]) or (
            not options["add"] and not options["clean"]
        ):
            raise CommandError('Please specify either "--add" or "--clean"')

        if not settings.COMPRESS_MTIME_DELAY:
            raise CommandError(
                "mtime caching is currently disabled. Please "
                "set the COMPRESS_MTIME_DELAY setting to a number of seconds."
            )

        files_to_add = set()
        keys_to_delete = set()

        for root, dirs, files in os.walk(
            settings.COMPRESS_ROOT, followlinks=options["follow_links"]
        ):
            for dir_ in dirs:
                if self.is_ignored(dir_):
                    dirs.remove(dir_)
            for filename in files:
                common = "".join(root.split(settings.COMPRESS_ROOT))
                if common.startswith(os.sep):
                    common = common[len(os.sep) :]
                if self.is_ignored(os.path.join(common, filename)):
                    continue
                filename = os.path.join(root, filename)
                keys_to_delete.add(get_mtime_cachekey(filename))
                if options["add"]:
                    files_to_add.add(filename)

        if keys_to_delete:
            cache.delete_many(list(keys_to_delete))
            self.stdout.write(
                "Deleted mtimes of %d files from the cache." % len(keys_to_delete)
            )

        if files_to_add:
            for filename in files_to_add:
                get_mtime(filename)
            self.stdout.write("Added mtimes of %d files to cache." % len(files_to_add))



import html
from importlib import import_module

from django.utils.functional import LazyObject

# support legacy parser module usage
from compressor.parser.base import ParserBase  # noqa
from compressor.parser.lxml import LxmlParser
from compressor.parser.default_htmlparser import DefaultHtmlParser as HtmlParser
from compressor.parser.beautifulsoup import BeautifulSoupParser  # noqa
from compressor.parser.html5lib import Html5LibParser  # noqa


class AutoSelectParser(LazyObject):
    options = (
        # TODO: make lxml.html parser first again
        (html.parser.__name__, HtmlParser),  # fast and part of the Python stdlib
        ("lxml.html", LxmlParser),  # lxml, extremely fast
    )

    def __init__(self, content):
        self._wrapped = None
        self._setup(content)

    def __getattr__(self, name):
        return getattr(self._wrapped, name)

    def _setup(self, content):
        for dependency, parser in self.options:
            try:
                import_module(dependency)
                self._wrapped = parser(content)
                break
            except (ImportError, TypeError):
                continue



from unittest.mock import Mock

from django.test import override_settings, TestCase

from compressor.css import CssCompressor
from compressor.js import JsCompressor
from compressor.signals import post_compress


@override_settings(
    COMPRESS_ENABLED=True, COMPRESS_PRECOMPILERS=(), COMPRESS_DEBUG_TOGGLE="nocompress"
)
class PostCompressSignalTestCase(TestCase):
    def setUp(self):
        self.css = """\
<link rel="stylesheet" href="/static/css/one.css" type="text/css">
<style type="text/css">p { border:5px solid green;}</style>
<link rel="stylesheet" href="/static/css/two.css" type="text/css">"""
        self.css_node = CssCompressor("css", self.css)

        self.js = """\
<script src="/static/js/one.js" type="text/javascript"></script>
<script type="text/javascript">obj.value = "value";</script>"""
        self.js_node = JsCompressor("js", self.js)

    def tearDown(self):
        post_compress.disconnect()

    def test_js_signal_sent(self):
        def listener(sender, **kwargs):
            pass

        callback = Mock(wraps=listener)
        post_compress.connect(callback)
        self.js_node.output()
        args, kwargs = callback.call_args
        self.assertEqual(JsCompressor, kwargs["sender"])
        self.assertEqual("js", kwargs["type"])
        self.assertEqual("file", kwargs["mode"])
        context = kwargs["context"]
        assert "url" in context["compressed"]

    def test_css_signal_sent(self):
        def listener(sender, **kwargs):
            pass

        callback = Mock(wraps=listener)
        post_compress.connect(callback)
        self.css_node.output()
        args, kwargs = callback.call_args
        self.assertEqual(CssCompressor, kwargs["sender"])
        self.assertEqual("css", kwargs["type"])
        self.assertEqual("file", kwargs["mode"])
        context = kwargs["context"]
        assert "url" in context["compressed"]

    def test_css_signal_multiple_media_attributes(self):
        css = """\
<link rel="stylesheet" href="/static/css/one.css" media="handheld" type="text/css" />
<style type="text/css" media="print">p { border:5px solid green;}</style>
<link rel="stylesheet" href="/static/css/two.css" type="text/css">"""
        css_node = CssCompressor("css", css)

        def listener(sender, **kwargs):
            pass

        callback = Mock(wraps=listener)
        post_compress.connect(callback)
        css_node.output()
        self.assertEqual(3, callback.call_count)



from django import template
from django.core.exceptions import ImproperlyConfigured

from compressor.cache import (
    cache_get,
    cache_set,
    get_offline_hexdigest,
    get_offline_manifest,
    get_templatetag_cachekey,
)
from compressor.conf import settings
from compressor.exceptions import OfflineGenerationError
from compressor.utils import get_class

register = template.Library()

OUTPUT_FILE = "file"
OUTPUT_INLINE = "inline"
OUTPUT_PRELOAD = "preload"
OUTPUT_MODES = (OUTPUT_FILE, OUTPUT_INLINE, OUTPUT_PRELOAD)


class CompressorMixin:
    def get_original_content(self, context):
        raise NotImplementedError

    @property
    def compressors(self):
        return settings.COMPRESSORS

    def compressor_cls(self, kind):
        if kind not in self.compressors.keys():
            raise template.TemplateSyntaxError(
                "The compress tag's argument must be one of: %s."
                % ", ".join(map(repr, self.compressors.keys()))
            )
        return get_class(self.compressors.get(kind), exception=ImproperlyConfigured)

    def get_compressor(self, context, kind, log, verbosity):
        cls = self.compressor_cls(kind)
        return cls(
            kind,
            content=self.get_original_content(context),
            context=context,
            log=log,
            verbosity=verbosity,
        )

    def debug_mode(self, context):
        if settings.COMPRESS_DEBUG_TOGGLE:
            # Only check for the debug parameter if a RequestContext was used
            request = context.get("request", None)
            if request is not None:
                return settings.COMPRESS_DEBUG_TOGGLE in request.GET

    def is_offline_compression_enabled(self, forced):
        """
        Check if offline compression is enabled or forced

        Defaults to just checking the settings and forced argument,
        but can be overridden to completely disable compression for
        a subclass, for instance.
        """
        return (settings.COMPRESS_ENABLED and settings.COMPRESS_OFFLINE) or forced

    def render_offline(self, context):
        """
        If enabled and in offline mode, and not forced check the offline cache
        and return the result if given
        """
        original_content = self.get_original_content(context)
        key = get_offline_hexdigest(original_content)
        offline_manifest = get_offline_manifest()
        if key in offline_manifest:
            return offline_manifest[key].replace(
                settings.COMPRESS_URL_PLACEHOLDER,
                # Cast ``settings.COMPRESS_URL`` to a string to allow it to be
                # a string-alike object to e.g. add ``SCRIPT_NAME`` WSGI param
                # as a *path prefix* to the output URL.
                # See https://code.djangoproject.com/ticket/25598.
                str(settings.COMPRESS_URL),
            )
        else:
            raise OfflineGenerationError(
                "You have offline compression "
                'enabled but key "%s" is missing from offline manifest. '
                'You may need to run "python manage.py compress". Here '
                "is the original content:\n\n%s" % (key, original_content)
            )

    def render_cached(self, compressor, kind, mode):
        """
        If enabled checks the cache for the given compressor's cache key
        and return a tuple of cache key and output
        """
        cache_key = get_templatetag_cachekey(compressor, mode, kind)
        cache_content = cache_get(cache_key)
        return cache_key, cache_content

    def render_compressed(
        self, context, kind, mode, name=None, forced=False, log=None, verbosity=0
    ):

        # See if it has been rendered offline
        if self.is_offline_compression_enabled(forced) and not forced:
            return self.render_offline(context)

        # Take a shortcut if we really don't have anything to do
        if (
            not settings.COMPRESS_ENABLED
            and not settings.COMPRESS_PRECOMPILERS
            and not forced
        ):
            return self.get_original_content(context)

        name = name or getattr(self, "name", None)
        context["compressed"] = {"name": name}
        compressor = self.get_compressor(context, kind, log, verbosity)

        # Check cache
        cache_key = None
        if settings.COMPRESS_ENABLED and not forced:
            cache_key, cache_content = self.render_cached(compressor, kind, mode)
            if cache_content is not None:
                return cache_content

        file_basename = name or getattr(self, "basename", None)
        if file_basename is None:
            file_basename = "output"

        rendered_output = compressor.output(mode, forced=forced, basename=file_basename)
        assert isinstance(rendered_output, str)
        if cache_key:
            cache_set(cache_key, rendered_output)
        return rendered_output


class CompressorNode(CompressorMixin, template.Node):
    def __init__(self, nodelist, kind=None, mode=OUTPUT_FILE, name=None):
        self.nodelist = nodelist
        self.kind = kind
        self.mode = mode
        self.name = name

    def get_original_content(self, context):
        return self.nodelist.render(context)

    def render(self, context, forced=False):

        # Check if in debug mode
        if self.debug_mode(context):
            return self.get_original_content(context)

        # pass logger to compressor object
        try:
            log, verbosity = context.template._log, context.template._log_verbosity
        except AttributeError:
            log, verbosity = None, 0

        return self.render_compressed(
            context, self.kind, self.mode, forced=forced, log=log, verbosity=verbosity
        )


@register.tag
def compress(parser, token):
    """
    Compresses linked and inline javascript or CSS into a single cached file.

    Syntax::

        {% compress <js/css> [<file/inline> [block_name]] %}
        <html of inline or linked JS/CSS>
        {% endcompress %}

    Examples::

        See docs/usage.rst

    """

    nodelist = parser.parse(("endcompress",))
    parser.delete_first_token()

    args = token.split_contents()

    if not len(args) in (2, 3, 4):
        raise template.TemplateSyntaxError(
            "%r tag requires either one, two or three arguments." % args[0]
        )

    kind = args[1]

    if len(args) >= 3:
        mode = args[2]
        if mode not in OUTPUT_MODES:
            raise template.TemplateSyntaxError(
                "%r's second argument must be '%s' or '%s'."
                % (args[0], OUTPUT_FILE, OUTPUT_INLINE)
            )
    else:
        mode = OUTPUT_FILE
    if len(args) == 4:
        name = args[3]
    else:
        name = None
    return CompressorNode(nodelist, kind, mode, name)



import os
import unittest

try:
    import lxml
except ImportError:
    lxml = None

try:
    import html5lib
except ImportError:
    html5lib = None

from django.test.utils import override_settings

from compressor.base import SOURCE_HUNK, SOURCE_FILE
from compressor.conf import settings
from compressor.tests.test_base import CompressorTestCase


class ParserTestCase:
    def setUp(self):
        self.override_settings = self.settings(COMPRESS_PARSER=self.parser_cls)
        self.override_settings.__enter__()
        super().setUp()

    def tearDown(self):
        self.override_settings.__exit__(None, None, None)


@unittest.skipIf(lxml is None, "lxml not found")
class LxmlParserTests(ParserTestCase, CompressorTestCase):
    parser_cls = "compressor.parser.LxmlParser"


@unittest.skipIf(html5lib is None, "html5lib not found")
class Html5LibParserTests(ParserTestCase, CompressorTestCase):
    parser_cls = "compressor.parser.Html5LibParser"
    # Special test variants required since xml.etree holds attributes
    # as a plain dictionary, e.g. key order is unpredictable.

    def test_css_split(self):
        split = self.css_node.split_contents()
        out0 = (
            SOURCE_FILE,
            os.path.join(settings.COMPRESS_ROOT, "css", "one.css"),
            "css/one.css",
            "{http://www.w3.org/1999/xhtml}link",
            {"rel": "stylesheet", "href": "/static/css/one.css", "type": "text/css"},
        )
        self.assertEqual(out0, split[0][:3] + (split[0][3].tag, split[0][3].attrib))
        out1 = (
            SOURCE_HUNK,
            "p { border:5px solid green;}",
            None,
            '<style type="text/css">p { border:5px solid green;}</style>',
        )
        self.assertEqual(
            out1, split[1][:3] + (self.css_node.parser.elem_str(split[1][3]),)
        )
        out2 = (
            SOURCE_FILE,
            os.path.join(settings.COMPRESS_ROOT, "css", "two.css"),
            "css/two.css",
            "{http://www.w3.org/1999/xhtml}link",
            {"rel": "stylesheet", "href": "/static/css/two.css", "type": "text/css"},
        )
        self.assertEqual(out2, split[2][:3] + (split[2][3].tag, split[2][3].attrib))

    def test_js_split(self):
        split = self.js_node.split_contents()
        out0 = (
            SOURCE_FILE,
            os.path.join(settings.COMPRESS_ROOT, "js", "one.js"),
            "js/one.js",
            "{http://www.w3.org/1999/xhtml}script",
            {"src": "/static/js/one.js", "type": "text/javascript"},
            None,
        )
        self.assertEqual(
            out0, split[0][:3] + (split[0][3].tag, split[0][3].attrib, split[0][3].text)
        )
        out1 = (
            SOURCE_HUNK,
            'obj.value = "value";',
            None,
            "{http://www.w3.org/1999/xhtml}script",
            {"type": "text/javascript"},
            'obj.value = "value";',
        )
        self.assertEqual(
            out1, split[1][:3] + (split[1][3].tag, split[1][3].attrib, split[1][3].text)
        )

    @override_settings(COMPRESS_ENABLED=False)
    def test_css_return_if_off(self):
        # Yes, they are semantically equal but attributes might be
        # scrambled in unpredictable order. A more elaborate check
        # would require parsing both arguments with a different parser
        # and then evaluating the result, which no longer is
        # a meaningful unit test.
        self.assertEqual(len(self.css), len(self.css_node.output()))

    @override_settings(COMPRESS_PRECOMPILERS=(), COMPRESS_ENABLED=False)
    def test_js_return_if_off(self):
        # As above.
        self.assertEqual(len(self.js), len(self.js_node.output()))


class BeautifulSoupParserTests(ParserTestCase, CompressorTestCase):
    parser_cls = "compressor.parser.BeautifulSoupParser"
    # just like in the Html5LibParserTests, provide special tests because
    # in bs4 attributes are held in dictionaries

    def test_css_split(self):
        split = self.css_node.split_contents()
        out0 = (
            SOURCE_FILE,
            os.path.join(settings.COMPRESS_ROOT, "css", "one.css"),
            "css/one.css",
            None,
            None,
        )
        self.assertEqual(out0, split[0][:3] + (split[0][3].tag, split[0][3].attrib))
        out1 = (
            SOURCE_HUNK,
            "p { border:5px solid green;}",
            None,
            '<style type="text/css">p { border:5px solid green;}</style>',
        )
        self.assertEqual(
            out1, split[1][:3] + (self.css_node.parser.elem_str(split[1][3]),)
        )
        out2 = (
            SOURCE_FILE,
            os.path.join(settings.COMPRESS_ROOT, "css", "two.css"),
            "css/two.css",
            None,
            None,
        )
        self.assertEqual(out2, split[2][:3] + (split[2][3].tag, split[2][3].attrib))

    @override_settings(COMPRESS_ENABLED=False)
    def test_css_return_if_off(self):
        self.assertEqual(len(self.css), len(self.css_node.output()))


class HtmlParserTests(ParserTestCase, CompressorTestCase):
    parser_cls = "compressor.parser.HtmlParser"



from django.core.exceptions import ImproperlyConfigured
from django.utils.encoding import smart_str

from compressor.parser import ParserBase


class BeautifulSoupParser(ParserBase):
    def __init__(self, content):
        super().__init__(content)
        try:
            from bs4 import BeautifulSoup

            # Disable multi_valued_attributes
            # http://www.crummy.com/software/BeautifulSoup/bs4/doc/#multi-valued-attributes
            self.soup = BeautifulSoup(self.content, "html.parser", multi_valued_attributes=None)
        except ImportError as err:
            raise ImproperlyConfigured("Error while importing BeautifulSoup: %s" % err)

    def css_elems(self):
        return self.soup.find_all({"link": True, "style": True})

    def js_elems(self):
        return self.soup.find_all("script")

    def elem_attribs(self, elem):
        return elem.attrs

    def elem_content(self, elem):
        return elem.string

    def elem_name(self, elem):
        return elem.name

    def elem_str(self, elem):
        elem_as_string = smart_str(elem)
        if elem.name == "link":
            # This makes testcases happy
            elem_as_string = elem_as_string.replace("/>", ">")
        return elem_as_string



from django.core.exceptions import ImproperlyConfigured
from django.utils.encoding import smart_str
from django.utils.functional import cached_property

from compressor.exceptions import ParserError
from compressor.parser import ParserBase


class LxmlParser(ParserBase):
    """
    LxmlParser will use `lxml.html` parser to parse rendered contents of
    {% compress %} tag.
    """

    def __init__(self, content):
        try:
            from lxml.html import fromstring
            from lxml.etree import tostring
        except ImportError as err:
            raise ImproperlyConfigured("Error while importing lxml: %s" % err)
        except Exception as err:
            raise ParserError("Error while initializing parser: %s" % err)

        self.fromstring = fromstring
        self.tostring = tostring
        super().__init__(content)

    @cached_property
    def tree(self):
        """
        Document tree.
        """
        content = "<root>%s</root>" % self.content
        tree = self.fromstring(content)
        self.tostring(tree, encoding=str)
        return tree

    def css_elems(self):
        return self.tree.xpath(
            '//link[re:test(@rel, "^stylesheet$", "i")]|style',
            namespaces={"re": "http://exslt.org/regular-expressions"},
        )

    def js_elems(self):
        return self.tree.findall("script")

    def elem_attribs(self, elem):
        return elem.attrib

    def elem_content(self, elem):
        return smart_str(elem.text)

    def elem_name(self, elem):
        return elem.tag

    def elem_str(self, elem):
        return smart_str(self.tostring(elem, method="html", encoding=str))



import html.parser

from django.utils.encoding import smart_str

from compressor.exceptions import ParserError
from compressor.parser import ParserBase


# The HTMLParser constructor takes a 'convert_charrefs'
# argument which raises a warning if we don't pass it.
HTML_PARSER_ARGS = {
    "convert_charrefs": False,
}


class DefaultHtmlParser(ParserBase, html.parser.HTMLParser):
    def __init__(self, content):
        html.parser.HTMLParser.__init__(self, **HTML_PARSER_ARGS)
        self.content = content
        self._css_elems = []
        self._js_elems = []
        self._current_tag = None
        try:
            self.feed(self.content)
            self.close()
        except Exception as err:
            lineno = err.lineno
            line = self.content.splitlines()[lineno]
            raise ParserError(
                "Error while initializing HtmlParser: %s (line: %s)" % (err, repr(line))
            )

    def handle_starttag(self, tag, attrs):
        tag = tag.lower()
        if tag in ("style", "script"):
            if tag == "style":
                tags = self._css_elems
            elif tag == "script":
                tags = self._js_elems
            tags.append(
                {"tag": tag, "attrs": attrs, "attrs_dict": dict(attrs), "text": ""}
            )
            self._current_tag = tag
        elif tag == "link":
            self._css_elems.append(
                {"tag": tag, "attrs": attrs, "attrs_dict": dict(attrs), "text": None}
            )

    def handle_endtag(self, tag):
        if self._current_tag and self._current_tag == tag.lower():
            self._current_tag = None

    def handle_data(self, data):
        if self._current_tag == "style":
            self._css_elems[-1]["text"] = data
        elif self._current_tag == "script":
            self._js_elems[-1]["text"] = data

    def css_elems(self):
        return self._css_elems

    def js_elems(self):
        return self._js_elems

    def elem_name(self, elem):
        return elem["tag"]

    def elem_attribs(self, elem):
        return elem["attrs_dict"]

    def elem_content(self, elem):
        return smart_str(elem["text"])

    def elem_str(self, elem):
        tag = {}
        tag.update(elem)
        tag["attrs"] = ""
        if len(elem["attrs"]):
            tag["attrs"] = " %s" % " ".join(
                ['%s="%s"' % (name, value) for name, value in elem["attrs"]]
            )
        if elem["tag"] == "link":
            return "<%(tag)s%(attrs)s>" % tag
        else:
            return "<%(tag)s%(attrs)s>%(text)s</%(tag)s>" % tag



import os
import codecs
from importlib import import_module
from urllib.request import url2pathname

from django.core.files.base import ContentFile
from django.utils.safestring import mark_safe
from django.template.loader import render_to_string
from django.utils.functional import cached_property

from compressor.cache import get_hexdigest, get_mtime
from compressor.conf import settings
from compressor.exceptions import (
    CompressorError,
    UncompressableFileError,
    FilterDoesNotExist,
)
from compressor.filters import CachedCompilerFilter
from compressor.storage import compressor_file_storage
from compressor.signals import post_compress
from compressor.utils import get_class, get_mod_func, staticfiles

# Some constants for nicer handling.
SOURCE_HUNK, SOURCE_FILE = "inline", "file"
METHOD_INPUT, METHOD_OUTPUT = "input", "output"


class Compressor:
    """
    Base compressor object to be subclassed for content type
    depending implementations details.
    """

    output_mimetypes = {}

    def __init__(
        self,
        resource_kind,
        content=None,
        output_prefix=None,
        context=None,
        filters=None,
        log=None,
        verbosity=1,
        *args,
        **kwargs
    ):
        if filters is None:
            self.filters = settings.COMPRESS_FILTERS[resource_kind]
        else:
            self.filters = filters
        if output_prefix is None:
            self.output_prefix = resource_kind
        else:
            self.output_prefix = output_prefix
        self.content = content or ""  # rendered contents of {% compress %} tag
        self.output_dir = settings.COMPRESS_OUTPUT_DIR.strip("/")
        self.charset = settings.DEFAULT_CHARSET
        self.split_content = []
        self.context = context or {}
        self.resource_kind = resource_kind
        self.extra_context = {}
        self.precompiler_mimetypes = dict(settings.COMPRESS_PRECOMPILERS)
        self.finders = staticfiles.finders
        self._storage = None
        self.log = log
        self.verbosity = verbosity

    def copy(self, **kwargs):
        keywords = dict(
            content=self.content,
            context=self.context,
            output_prefix=self.output_prefix,
            filters=self.filters,
        )
        keywords.update(kwargs)
        return self.__class__(self.resource_kind, **keywords)

    @cached_property
    def storage(self):
        from compressor.storage import default_storage

        return default_storage

    def split_contents(self):
        """
        To be implemented in a subclass, should return an
        iterable with four values: kind, value, basename, element
        """
        raise NotImplementedError

    def get_template_name(self, mode):
        """
        Returns the template path for the given mode.
        """
        try:
            template = getattr(self, "template_name_%s" % mode)
            if template:
                return template
        except AttributeError:
            pass
        return "compressor/%s_%s.html" % (self.resource_kind, mode)

    def get_basename(self, url):
        """
        Takes full path to a static file (eg. "/static/css/style.css") and
        returns path with storage's base url removed (eg. "css/style.css").
        """
        try:
            base_url = self.storage.base_url
        except AttributeError:
            base_url = settings.COMPRESS_URL

        # Cast ``base_url`` to a string to allow it to be
        # a string-alike object to e.g. add ``SCRIPT_NAME``
        # WSGI param as a *path prefix* to the output URL.
        # See https://code.djangoproject.com/ticket/25598.
        base_url = str(base_url)

        if not url.startswith(base_url):
            raise UncompressableFileError(
                "'%s' isn't accessible via "
                "COMPRESS_URL ('%s') and can't be "
                "compressed" % (url, base_url)
            )
        basename = url.replace(base_url, "", 1)
        # drop the querystring, which is used for non-compressed cache-busting.
        return basename.split("?", 1)[0]

    def get_filepath(self, content, basename=None):
        """
        Returns file path for an output file based on contents.

        Returned path is relative to compressor storage's base url, for
        example "CACHE/css/58a8c0714e59.css".

        When `basename` argument is provided then file name (without extension)
        will be used as a part of returned file name, for example:

        get_filepath(content, "my_file.css") -> 'CACHE/css/my_file.58a8c0714e59.css'
        """
        parts = []
        if basename:
            filename = os.path.split(basename)[1]
            parts.append(os.path.splitext(filename)[0])
        parts.extend([get_hexdigest(content, 12), self.resource_kind])
        return os.path.join(self.output_dir, self.output_prefix, ".".join(parts))

    def get_filename(self, basename):
        """
        Returns full path to a file, for example:

        get_filename('css/one.css') -> '/full/path/to/static/css/one.css'
        """
        filename = None
        # First try finding the file using the storage class.
        # This is skipped in DEBUG mode as files might be outdated in
        # compressor's final destination (COMPRESS_ROOT) during development
        if not settings.DEBUG:
            try:
                # call path first so remote storages don't make it to exists,
                # which would cause network I/O
                if self.log and self.verbosity >= 2:
                    self.log.write("Looking for '{}' in storage\n".format(basename))
                filename = self.storage.path(basename)
                if not self.storage.exists(basename):
                    filename = None
            except NotImplementedError:
                # remote storages don't implement path, access the file locally
                if self.log and self.verbosity >= 2:
                    self.log.write(
                        "Remote storages don't implement path, looking for the file locally\n"
                    )
                if compressor_file_storage.exists(basename):
                    filename = compressor_file_storage.path(basename)
        # secondly try to find it with staticfiles
        if not filename and self.finders:
            if self.log and self.verbosity >= 2:
                if not settings.DEBUG:
                    self.log.write(
                        "'{}' was not found in storage, using static finders\n".format(
                            basename
                        )
                    )
                else:
                    self.log.write("Using static finders for '{}'\n".format(basename))
            filename = self.finders.find(url2pathname(basename))
        if filename:
            return filename
        # or just raise an exception as the last resort
        raise UncompressableFileError(
            "'%s' could not be found in the COMPRESS_ROOT '%s'%s"
            % (
                basename,
                settings.COMPRESS_ROOT,
                self.finders and " or with staticfiles." or ".",
            )
        )

    def get_filecontent(self, filename, charset):
        """
        Reads file contents using given `charset` and returns it as text.
        """
        if charset == "utf-8":
            # Removes BOM
            charset = "utf-8-sig"
        with codecs.open(filename, "r", charset) as fd:
            try:
                return fd.read()
            except IOError as e:
                raise UncompressableFileError(
                    "IOError while processing " "'%s': %s" % (filename, e)
                )
            except UnicodeDecodeError as e:
                raise UncompressableFileError(
                    "UnicodeDecodeError while "
                    "processing '%s' with "
                    "charset %s: %s" % (filename, charset, e)
                )

    @cached_property
    def parser(self):
        return get_class(settings.COMPRESS_PARSER)(self.content)

    @cached_property
    def cached_filters(self):
        return [get_class(filter_cls) for filter_cls in self.filters]

    @cached_property
    def mtimes(self):
        return [
            str(get_mtime(value))
            for kind, value, basename, elem in self.split_contents()
            if kind == SOURCE_FILE
        ]

    @cached_property
    def cachekey(self):
        return get_hexdigest(
            "".join([self.content] + self.mtimes).encode(self.charset), 12
        )

    def hunks(self, forced=False):
        """
        The heart of content parsing, iterates over the
        list of split contents and looks at its kind
        to decide what to do with it. Should yield a
        bunch of precompiled and/or rendered hunks.
        """
        enabled = settings.COMPRESS_ENABLED or forced

        for kind, value, basename, elem in self.split_contents():
            precompiled = False
            attribs = self.parser.elem_attribs(elem)
            charset = attribs.get("charset", self.charset)
            options = {
                "method": METHOD_INPUT,
                "elem": elem,
                "kind": kind,
                "basename": basename,
                "charset": charset,
            }

            if kind == SOURCE_FILE:
                options = dict(options, filename=value)
                value = self.get_filecontent(value, charset)

            if self.precompiler_mimetypes:
                precompiled, value = self.precompile(value, **options)

            if enabled:
                yield self.filter(value, self.cached_filters, **options)
            elif precompiled:
                for filter_cls in self.cached_filters:
                    if filter_cls.run_with_compression_disabled:
                        value = self.filter(value, [filter_cls], **options)
                yield self.handle_output(kind, value, forced=True, basename=basename)
            else:
                yield self.parser.elem_str(elem)

    def filter_output(self, content):
        """
        Passes the concatenated content to the 'output' methods
        of the compressor filters.
        """
        return self.filter(content, self.cached_filters, method=METHOD_OUTPUT)

    def filter_input(self, forced=False):
        """
        Passes each hunk (file or code) to the 'input' methods
        of the compressor filters.
        """
        content = []
        for hunk in self.hunks(forced):
            content.append(hunk)
        return content

    def precompile(
        self, content, kind=None, elem=None, filename=None, charset=None, **kwargs
    ):
        """
        Processes file using a pre compiler.

        This is the place where files like coffee script are processed.
        """
        if not kind:
            return False, content
        attrs = self.parser.elem_attribs(elem)
        mimetype = attrs.get("type", None)
        if mimetype is None:
            return False, content

        filter_or_command = self.precompiler_mimetypes.get(mimetype)
        if filter_or_command is None:
            if mimetype in self.output_mimetypes:
                return False, content
            raise CompressorError(
                "Couldn't find any precompiler in "
                "COMPRESS_PRECOMPILERS setting for "
                "mimetype '%s'." % mimetype
            )

        mod_name, cls_name = get_mod_func(filter_or_command)
        try:
            mod = import_module(mod_name)
        except (ImportError, TypeError):
            filter = CachedCompilerFilter(
                content=content,
                filter_type=self.resource_kind,
                filename=filename,
                charset=charset,
                command=filter_or_command,
                mimetype=mimetype,
            )
            return True, filter.input(**kwargs)
        try:
            precompiler_class = getattr(mod, cls_name)
        except AttributeError:
            raise FilterDoesNotExist('Could not find "%s".' % filter_or_command)
        filter = precompiler_class(
            content,
            attrs=attrs,
            filter_type=self.resource_kind,
            charset=charset,
            filename=filename,
        )
        return True, filter.input(**kwargs)

    def filter(self, content, filters, method, **kwargs):
        for filter_cls in filters:
            filter_func = getattr(
                filter_cls(content, filter_type=self.resource_kind), method
            )
            try:
                if callable(filter_func):
                    content = filter_func(**kwargs)
            except NotImplementedError:
                pass
        return content

    def output(self, mode="file", forced=False, basename=None):
        """
        The general output method, override in subclass if you need to do
        any custom modification. Calls other mode specific methods or simply
        returns the content directly.
        """
        output = "\n".join(self.filter_input(forced))

        if not output:
            return ""

        if settings.COMPRESS_ENABLED or forced:
            filtered_output = self.filter_output(output)
            return self.handle_output(mode, filtered_output, forced, basename)

        return output

    def handle_output(self, mode, content, forced, basename=None):
        # Then check for the appropriate output method and call it
        output_func = getattr(self, "output_%s" % mode, None)
        if callable(output_func):
            return output_func(mode, content, forced, basename)
        # Total failure, raise a general exception
        raise CompressorError("Couldn't find output method for mode '%s'" % mode)

    def output_file(self, mode, content, forced=False, basename=None):
        """
        The output method that saves the content to a file and renders
        the appropriate template with the file's URL.
        """
        new_filepath = self.get_filepath(content, basename=basename)
        if not self.storage.exists(new_filepath) or forced:
            self.storage.save(new_filepath, ContentFile(content.encode(self.charset)))
        url = mark_safe(self.storage.url(new_filepath))
        return self.render_output(mode, {"url": url})

    def output_inline(self, mode, content, forced=False, basename=None):
        """
        The output method that directly returns the content for inline
        display.
        """
        return self.render_output(mode, {"content": content})

    def output_preload(self, mode, content, forced=False, basename=None):
        """
        The output method that returns <link> with rel="preload" and
        proper href attribute for given file.
        """
        return self.output_file(mode, content, forced, basename)

    def render_output(self, mode, context=None):
        """
        Renders the compressor output with the appropriate template for
        the given mode and template context.
        """
        # Just in case someone renders the compressor outside
        # the usual template rendering cycle
        if "compressed" not in self.context:
            self.context["compressed"] = {}

        self.context["compressed"].update(context or {})
        self.context["compressed"].update(self.extra_context)

        if hasattr(self.context, "flatten"):
            # Passing Contexts to Template.render is deprecated since Django 1.8.
            final_context = self.context.flatten()
        else:
            final_context = self.context

        post_compress.send(
            sender=self.__class__,
            type=self.resource_kind,
            mode=mode,
            context=final_context,
        )
        template_name = self.get_template_name(mode)
        return render_to_string(template_name, context=final_context)



import gzip
import os
from datetime import datetime
import time
from urllib.parse import urljoin

from django.core.files.storage import FileSystemStorage, InvalidStorageError, storages
from django.utils.functional import LazyObject, SimpleLazyObject

from compressor.conf import settings


def get_storage(alias=None, storage_class=None):
    alias = alias or settings.COMPRESS_STORAGE_ALIAS
    storage_class = storage_class or settings.COMPRESS_STORAGE

    try:
        return storages[alias]
    except InvalidStorageError:
        storages.backends[alias] = {
            "BACKEND": storage_class
        }
        return storages[alias]


class CompressorFileStorage(FileSystemStorage):
    """
    Standard file system storage for files handled by django-compressor.

    The defaults for ``location`` and ``base_url`` are ``COMPRESS_ROOT`` and
    ``COMPRESS_URL``.

    """

    def __init__(self, location=None, base_url=None, *args, **kwargs):
        if location is None:
            location = settings.COMPRESS_ROOT
        if base_url is None:
            base_url = settings.COMPRESS_URL
        super().__init__(location, base_url, *args, **kwargs)

    def accessed_time(self, name):
        return datetime.fromtimestamp(os.path.getatime(self.path(name)))

    def created_time(self, name):
        return datetime.fromtimestamp(os.path.getctime(self.path(name)))

    def modified_time(self, name):
        return datetime.fromtimestamp(os.path.getmtime(self.path(name)))

    def save(self, filename, content):
        temp_filename = super().save(filename, content)
        # If a file already exists  in the target location, FileSystemStorage
        # will generate an unique filename and save content there instead.
        # When that happens, we move the file to the intended location using
        # os.replace() (which is an atomic operation):
        if temp_filename != filename:
            os.replace(self.path(temp_filename), self.path(filename))

        return filename


compressor_file_storage = SimpleLazyObject(
    lambda: storages.create_storage(
        {"BACKEND": "compressor.storage.CompressorFileStorage"}
    )
)


class GzipCompressorFileStorage(CompressorFileStorage):
    """
    File system storage that stores gzipped files in addition to the usual files.
    """

    def save(self, filename, content):
        filename = super().save(filename, content)
        orig_path = self.path(filename)
        compressed_path = "%s.gz" % orig_path

        with open(orig_path, "rb") as f_in, open(compressed_path, "wb") as f_out:
            with gzip.GzipFile(fileobj=f_out, mode="wb") as gz_out:
                gz_out.write(f_in.read())

        # Ensure the file timestamps match.
        # os.stat() returns nanosecond resolution on Linux, but os.utime()
        # only sets microsecond resolution.  Set times on both files to
        # ensure they are equal.
        stamp = time.time()
        os.utime(orig_path, (stamp, stamp))
        os.utime(compressed_path, (stamp, stamp))

        return filename


class BrotliCompressorFileStorage(CompressorFileStorage):
    """
    File system storage that stores brotli files in addition to the usual files.
    """

    chunk_size = 1024

    def save(self, filename, content):
        filename = super().save(filename, content)
        orig_path = self.path(filename)
        compressed_path = "%s.br" % orig_path

        import brotli

        br_compressor = brotli.Compressor()
        with open(orig_path, "rb") as f_in, open(compressed_path, "wb") as f_out:
            for f_in_data in iter(lambda: f_in.read(self.chunk_size), b""):
                compressed_data = br_compressor.process(f_in_data)
                if not compressed_data:
                    compressed_data = br_compressor.flush()
                f_out.write(compressed_data)
            f_out.write(br_compressor.finish())
        # Ensure the file timestamps match.
        # os.stat() returns nanosecond resolution on Linux, but os.utime()
        # only sets microsecond resolution.  Set times on both files to
        # ensure they are equal.
        stamp = time.time()
        os.utime(orig_path, (stamp, stamp))
        os.utime(compressed_path, (stamp, stamp))

        return filename


class DefaultStorage(LazyObject):
    def _setup(self):
        self._wrapped = get_storage()


default_storage = DefaultStorage()


class OfflineManifestFileStorage(CompressorFileStorage):
    def __init__(self, location=None, base_url=None, *args, **kwargs):
        if location is None:
            location = os.path.join(
                settings.COMPRESS_ROOT, settings.COMPRESS_OUTPUT_DIR
            )
        if base_url is None:
            base_url = urljoin(settings.COMPRESS_URL, settings.COMPRESS_OUTPUT_DIR)
        super().__init__(location, base_url, *args, **kwargs)


class DefaultOfflineManifestStorage(LazyObject):
    def _setup(self):
        self._wrapped = get_storage(
            alias=settings.COMPRESS_OFFLINE_MANIFEST_STORAGE_ALIAS,
            storage_class=settings.COMPRESS_OFFLINE_MANIFEST_STORAGE,
        )


default_offline_manifest_storage = DefaultOfflineManifestStorage()



import os
import re
import sys
from tempfile import mkdtemp
from shutil import rmtree, copytree

from bs4 import BeautifulSoup

from django.core.cache.backends import locmem
from django.test import SimpleTestCase
from django.test.utils import override_settings

from compressor import cache as cachemod
from compressor.base import SOURCE_FILE, SOURCE_HUNK
from compressor.cache import get_cachekey, get_precompiler_cachekey
from compressor.conf import settings
from compressor.css import CssCompressor
from compressor.exceptions import FilterDoesNotExist, FilterError
from compressor.js import JsCompressor
from compressor.storage import DefaultStorage


def make_soup(markup):
    return BeautifulSoup(markup, "html.parser")


def css_tag(href, **kwargs):
    rendered_attrs = "".join([' %s="%s"' % (k, v) for k, v in kwargs.items()])
    template = '<link rel="stylesheet" href="%s" type="text/css"%s>'
    return template % (href, rendered_attrs)


class TestPrecompiler:
    """A filter whose output is always the string 'OUTPUT'"""

    def __init__(self, content, attrs, filter_type=None, filename=None, charset=None):
        pass

    def input(self, **kwargs):
        return "OUTPUT"


class PassthroughPrecompiler:
    """A filter whose outputs the input unmodified"""

    def __init__(self, content, attrs, filter_type=None, filename=None, charset=None):
        self.content = content

    def input(self, **kwargs):
        return self.content


test_dir = os.path.abspath(os.path.join(os.path.dirname(__file__)))


class PrecompilerAndAbsoluteFilterTestCase(SimpleTestCase):
    def setUp(self):
        self.html_orig = '<link rel="stylesheet" href="/static/css/relative_url.css" type="text/css" />'
        self.html_auto_close_removed = '<link rel="stylesheet" href="/static/css/relative_url.css" type="text/css">'
        self.html_link_to_precompiled_css = '<link rel="stylesheet" href="/static/CACHE/css/relative_url.e8602322bfa6.css" type="text/css">'
        self.html_link_to_absolutized_css = '<link rel="stylesheet" href="/static/CACHE/css/relative_url.376db5682982.css" type="text/css">'
        self.css_orig = (
            "p { background: url('../img/python.png'); }"  # content of relative_url.css
        )
        self.css_absolutized = (
            "p { background: url('/static/img/python.png?ccb38978f900'); }"
        )

    def helper(self, enabled, use_precompiler, use_absolute_filter, expected_output):
        precompiler = (
            (("text/css", "compressor.tests.test_base.PassthroughPrecompiler"),)
            if use_precompiler
            else ()
        )
        filters = (
            ("compressor.filters.css_default.CssAbsoluteFilter",)
            if use_absolute_filter
            else ()
        )

        with self.settings(
            COMPRESS_ENABLED=enabled,
            COMPRESS_PRECOMPILERS=precompiler,
            COMPRESS_FILTERS={"css": filters},
        ):
            css_node = CssCompressor("css", self.html_orig)
            output = list(css_node.hunks())[0]
            self.assertEqual(output, expected_output)

    @override_settings(COMPRESS_CSS_HASHING_METHOD="content")
    def test_precompiler_enables_absolute(self):
        """
        Tests whether specifying a precompiler also runs the CssAbsoluteFilter even if
        compression is disabled, but only if the CssAbsoluteFilter is actually contained
        in the filters setting.
        While at it, ensure that everything runs as expected when compression is enabled.
        """
        self.helper(
            enabled=False,
            use_precompiler=False,
            use_absolute_filter=False,
            expected_output=self.html_auto_close_removed,
        )
        self.helper(
            enabled=False,
            use_precompiler=False,
            use_absolute_filter=True,
            expected_output=self.html_auto_close_removed,
        )
        self.helper(
            enabled=False,
            use_precompiler=True,
            use_absolute_filter=False,
            expected_output=self.html_link_to_precompiled_css,
        )
        self.helper(
            enabled=False,
            use_precompiler=True,
            use_absolute_filter=True,
            expected_output=self.html_link_to_absolutized_css,
        )
        self.helper(
            enabled=True,
            use_precompiler=False,
            use_absolute_filter=False,
            expected_output=self.css_orig,
        )
        self.helper(
            enabled=True,
            use_precompiler=False,
            use_absolute_filter=True,
            expected_output=self.css_absolutized,
        )
        self.helper(
            enabled=True,
            use_precompiler=True,
            use_absolute_filter=False,
            expected_output=self.css_orig,
        )
        self.helper(
            enabled=True,
            use_precompiler=True,
            use_absolute_filter=True,
            expected_output=self.css_absolutized,
        )


@override_settings(
    COMPRESS_ENABLED=True,
    COMPRESS_PRECOMPILERS=(),
    COMPRESS_DEBUG_TOGGLE="nocompress",
)
class CompressorTestCase(SimpleTestCase):
    def setUp(self):
        self.css = """\
<link rel="stylesheet" href="/static/css/one.css" type="text/css">
<style type="text/css">p { border:5px solid green;}</style>
<link rel="stylesheet" href="/static/css/two.css" type="text/css">"""
        self.css_node = CssCompressor("css", self.css)

        self.js = """\
<script src="/static/js/one.js" type="text/javascript"></script>
<script type="text/javascript">obj.value = "value";</script>"""
        self.js_node = JsCompressor("js", self.js)

    def assertEqualCollapsed(self, a, b):
        """
        assertEqual with internal newlines collapsed to single, and
        trailing whitespace removed.
        """

        def collapse(s):
            return re.sub(r"\n+", "\n", s).rstrip()

        self.assertEqual(collapse(a), collapse(b))

    def assertEqualSplits(self, a, b):
        """
        assertEqual for splits, particularly ignoring the presence of
        a trailing newline on the content.
        """

        def mangle(split):
            return [(x[0], x[1], x[2], x[3].rstrip()) for x in split]

        self.assertEqual(mangle(a), mangle(b))

    def test_css_split(self):
        out = [
            (
                SOURCE_FILE,
                os.path.join(settings.COMPRESS_ROOT, "css", "one.css"),
                "css/one.css",
                '<link rel="stylesheet" href="/static/css/one.css" type="text/css">',
            ),
            (
                SOURCE_HUNK,
                "p { border:5px solid green;}",
                None,
                '<style type="text/css">p { border:5px solid green;}</style>',
            ),
            (
                SOURCE_FILE,
                os.path.join(settings.COMPRESS_ROOT, "css", "two.css"),
                "css/two.css",
                '<link rel="stylesheet" href="/static/css/two.css" type="text/css">',
            ),
        ]
        split = self.css_node.split_contents()
        split = [(x[0], x[1], x[2], self.css_node.parser.elem_str(x[3])) for x in split]
        self.assertEqualSplits(split, out)

    def test_css_hunks(self):
        out = [
            "body { background:#990; }",
            "p { border:5px solid green;}",
            "body { color:#fff; }",
        ]
        self.assertEqual(out, list(self.css_node.hunks()))

    def test_css_output(self):
        out = "body { background:#990; }\np { border:5px solid green;}\nbody { color:#fff; }"
        hunks = "\n".join([h for h in self.css_node.hunks()])
        self.assertEqual(out, hunks)

    def test_css_output_with_bom_input(self):
        out = "body { background:#990; }\n.compress-test {color: red;}"
        css = """<link rel="stylesheet" href="/static/css/one.css" type="text/css" />
        <link rel="stylesheet" href="/static/css/utf-8_with-BOM.css" type="text/css" />"""
        css_node_with_bom = CssCompressor("css", css)
        hunks = "\n".join([h for h in css_node_with_bom.hunks()])
        self.assertEqual(out, hunks)

    def test_css_mtimes(self):
        is_date = re.compile(r"^\d{10}[\.\d]+$")
        for date in self.css_node.mtimes:
            self.assertTrue(
                is_date.match(str(float(date))),
                "mtimes is returning something that doesn't look like a date: %s"
                % date,
            )

    @override_settings(COMPRESS_ENABLED=False)
    def test_css_return_if_off(self):
        self.assertEqualCollapsed(self.css, self.css_node.output())

    def test_cachekey(self):
        is_cachekey = re.compile(r"\w{12}")
        self.assertTrue(
            is_cachekey.match(self.css_node.cachekey),
            r"cachekey is returning something that doesn't look like r'\w{12}'",
        )

    def test_css_return_if_on(self):
        output = css_tag("/static/CACHE/css/600674ea1d3d.css")
        self.assertEqual(output, self.css_node.output().strip())

    def test_css_preload_output(self):
        # this needs to have the same hash as in the test above
        out = '<link rel="preload" href="/static/CACHE/css/600674ea1d3d.css" as="style" />'
        self.assertEqual(out, self.css_node.output(mode="preload"))

    def test_js_split(self):
        out = [
            (
                SOURCE_FILE,
                os.path.join(settings.COMPRESS_ROOT, "js", "one.js"),
                "js/one.js",
                '<script src="/static/js/one.js" type="text/javascript"></script>',
            ),
            (
                SOURCE_HUNK,
                'obj.value = "value";',
                None,
                '<script type="text/javascript">obj.value = "value";</script>',
            ),
        ]
        split = self.js_node.split_contents()
        split = [(x[0], x[1], x[2], self.js_node.parser.elem_str(x[3])) for x in split]
        self.assertEqualSplits(split, out)

    def test_js_hunks(self):
        out = ["obj = {};", 'obj.value = "value";']
        self.assertEqual(out, list(self.js_node.hunks()))

    def test_js_output(self):
        out = '<script src="/static/CACHE/js/8a0fed36c317.js"></script>'
        self.assertEqual(out, self.js_node.output())

    def test_js_preload_output(self):
        # this needs to have the same hash as in the test above
        out = (
            '<link rel="preload" href="/static/CACHE/js/8a0fed36c317.js" as="script" />'
        )
        self.assertEqual(out, self.js_node.output(mode="preload"))

    def test_js_override_url(self):
        self.js_node.context.update({"url": "This is not a url, just a text"})
        out = '<script src="/static/CACHE/js/8a0fed36c317.js"></script>'
        self.assertEqual(out, self.js_node.output())

    def test_css_override_url(self):
        self.css_node.context.update({"url": "This is not a url, just a text"})
        output = css_tag("/static/CACHE/css/600674ea1d3d.css")
        self.assertEqual(output, self.css_node.output().strip())

    @override_settings(COMPRESS_PRECOMPILERS=(), COMPRESS_ENABLED=False)
    def test_js_return_if_off(self):
        self.assertEqualCollapsed(self.js, self.js_node.output())

    def test_js_return_if_on(self):
        output = '<script src="/static/CACHE/js/8a0fed36c317.js"></script>'
        self.assertEqual(output, self.js_node.output())

    @override_settings(COMPRESS_OUTPUT_DIR="custom")
    def test_custom_output_dir1(self):
        output = '<script src="/static/custom/js/8a0fed36c317.js"></script>'
        self.assertEqual(output, JsCompressor("js", self.js).output())

    @override_settings(COMPRESS_OUTPUT_DIR="")
    def test_custom_output_dir2(self):
        output = '<script src="/static/js/8a0fed36c317.js"></script>'
        self.assertEqual(output, JsCompressor("js", self.js).output())

    @override_settings(COMPRESS_OUTPUT_DIR="/custom/nested/")
    def test_custom_output_dir3(self):
        output = '<script src="/static/custom/nested/js/8a0fed36c317.js"></script>'
        self.assertEqual(output, JsCompressor("js", self.js).output())

    @override_settings(
        COMPRESS_PRECOMPILERS=(
            ("text/foobar", "compressor.tests.test_base.TestPrecompiler"),
        ),
        COMPRESS_ENABLED=True,
    )
    def test_precompiler_class_used(self):
        css = '<style type="text/foobar">p { border:10px solid red;}</style>'
        css_node = CssCompressor("css", css)
        output = make_soup(css_node.output("inline"))
        self.assertEqual(output.style.contents[0], "OUTPUT")

    @override_settings(
        COMPRESS_PRECOMPILERS=(
            ("text/foobar", "compressor.tests.test_base.NonexistentFilter"),
        ),
        COMPRESS_ENABLED=True,
    )
    def test_nonexistent_precompiler_class_error(self):
        css = '<style type="text/foobar">p { border:10px solid red;}</style>'
        css_node = CssCompressor("css", css)
        self.assertRaises(FilterDoesNotExist, css_node.output, "inline")

    @override_settings(
        COMPRESS_PRECOMPILERS=(("text/foobar", "./foo -I ./bar/baz"),),
        COMPRESS_ENABLED=True,
    )
    def test_command_with_dot_precompiler(self):
        css = '<style type="text/foobar">p { border:10px solid red;}</style>'
        css_node = CssCompressor("css", css)
        self.assertRaises(FilterError, css_node.output, "inline")

    @override_settings(
        COMPRESS_PRECOMPILERS=(
            ("text/django", "compressor.filters.template.TemplateFilter"),
        ),
        COMPRESS_ENABLED=True,
    )
    def test_template_precompiler(self):
        css = '<style type="text/django">p { border:10px solid {% if 1 %}green{% else %}red{% endif %};}</style>'
        css_node = CssCompressor("css", css)
        output = make_soup(css_node.output("inline"))
        self.assertEqual(output.style.contents[0], "p{border:10px solid green}")


class CssMediaTestCase(SimpleTestCase):
    def setUp(self):
        self.css = """\
<link rel="stylesheet" href="/static/css/one.css" type="text/css" media="screen">
<style type="text/css" media="print">p { border:5px solid green;}</style>
<link rel="stylesheet" href="/static/css/two.css" type="text/css" media="all">
<style type="text/css">h1 { border:5px solid green;}</style>"""

    def test_css_output(self):
        css_node = CssCompressor("css", self.css)
        links = make_soup(css_node.output()).find_all("link")
        media = ["screen", "print", "all", None]
        self.assertEqual(len(links), 4)
        self.assertEqual(media, [link.get("media", None) for link in links])

    def test_avoid_reordering_css(self):
        css = (
            self.css
            + '<style type="text/css" media="print">p { border:10px solid red;}</style>'
        )
        css_node = CssCompressor("css", css)
        media = ["screen", "print", "all", None, "print"]
        links = make_soup(css_node.output()).find_all("link")
        self.assertEqual(media, [link.get("media", None) for link in links])

    @override_settings(
        COMPRESS_PRECOMPILERS=(
            (
                "text/foobar",
                "%s %s {infile} {outfile}"
                % (sys.executable, os.path.join(test_dir, "precompiler.py")),
            ),
        ),
        COMPRESS_ENABLED=False,
    )
    def test_passthough_when_compress_disabled(self):
        css = """\
<link rel="stylesheet" href="/static/css/one.css" type="text/css" media="screen">
<link rel="stylesheet" href="/static/css/two.css" type="text/css" media="screen">
<style type="text/foobar" media="screen">h1 { border:5px solid green;}</style>"""
        css_node = CssCompressor("css", css)
        output = make_soup(css_node.output()).find_all(["link", "style"])
        self.assertEqual(
            ["/static/css/one.css", "/static/css/two.css", None],
            [link.get("href", None) for link in output],
        )
        self.assertEqual(
            ["screen", "screen", "screen"], [link.get("media", None) for link in output]
        )


@override_settings(COMPRESS_VERBOSE=True)
class VerboseTestCase(CompressorTestCase):
    pass


class CacheBackendTestCase(CompressorTestCase):
    def test_correct_backend(self):
        from compressor.cache import cache

        self.assertEqual(cache.__class__, locmem.LocMemCache)


class JsAsyncDeferTestCase(SimpleTestCase):
    def setUp(self):
        self.js = """\
            <script src="/static/js/one.js" type="text/javascript"></script>
            <script src="/static/js/two.js" type="text/javascript" async></script>
            <script src="/static/js/three.js" type="text/javascript" defer></script>
            <script type="text/javascript">obj.value = "value";</script>
            <script src="/static/js/one.js" type="text/javascript" async></script>
            <script src="/static/js/two.js" type="text/javascript" async></script>
            <script src="/static/js/three.js" type="text/javascript"></script>"""

    def test_js_output(self):
        def extract_attr(tag):
            if tag.has_attr("async"):
                return "async"
            if tag.has_attr("defer"):
                return "defer"

        js_node = JsCompressor("js", self.js)
        output = [None, "async", "defer", None, "async", None]
        scripts = make_soup(js_node.output()).find_all("script")
        attrs = [extract_attr(s) for s in scripts]
        self.assertEqual(output, attrs)


class JSWithParensTestCase(SimpleTestCase):
    def setUp(self):
        self.js = """
        <script src="/static/js/one.js"></script>
        <script src="/static/js/two.js"></script>
        """

    def test_js_content(self):
        js_node = JsCompressor("js", self.js)

        content = js_node.filter_input()
        self.assertEqual(content[0], "obj = {};;")
        self.assertEqual(content[1], "pollos = {};")


class CacheTestCase(SimpleTestCase):
    def setUp(self):
        cachemod._cachekey_func = None

    def test_get_cachekey_basic(self):
        self.assertEqual(get_cachekey("foo"), "django_compressor.foo")

    @override_settings(COMPRESS_CACHE_KEY_FUNCTION=".leading.dot")
    def test_get_cachekey_leading_dot(self):
        self.assertRaises(ImportError, lambda: get_cachekey("foo"))

    @override_settings(COMPRESS_CACHE_KEY_FUNCTION="invalid.module")
    def test_get_cachekey_invalid_mod(self):
        self.assertRaises(ImportError, lambda: get_cachekey("foo"))

    def test_get_precompiler_cachekey(self):
        try:
            get_precompiler_cachekey("asdf", "asdf")
        except TypeError:
            self.fail("get_precompiler_cachekey raised TypeError unexpectedly")


class CompressorInDebugModeTestCase(SimpleTestCase):
    def setUp(self):
        self.css = (
            '<link rel="stylesheet" href="/static/css/one.css" type="text/css" />'
        )
        self.expected_css_hash = "5c6a60375256"
        self.tmpdir = mkdtemp()
        new_static_root = os.path.join(self.tmpdir, "static")
        copytree(settings.STATIC_ROOT, new_static_root)

        self.override_settings = self.settings(
            COMPRESS_ENABLED=True,
            COMPRESS_PRECOMPILERS=(),
            COMPRESS_DEBUG_TOGGLE="nocompress",
            DEBUG=True,
            STATIC_ROOT=new_static_root,
            COMPRESS_ROOT=new_static_root,
            STATICFILES_DIRS=[settings.COMPRESS_ROOT],
        )
        self.override_settings.__enter__()

    def tearDown(self):
        rmtree(self.tmpdir)
        self.override_settings.__exit__(None, None, None)

    def test_filename_in_debug_mode(self):
        # In debug mode, compressor should look for files using staticfiles
        # finders only, and not look into the global static directory, where
        # files can be outdated. So compressor's output shouldn't change from
        # the one pre-generated if we modify the file in STATIC_ROOT.
        def compare():
            expected = (
                '<link rel="stylesheet" href="/static/CACHE/css/%s.css" type="text/css">'
                % self.expected_css_hash
            )
            compressor = CssCompressor("css", self.css)
            compressor.storage = DefaultStorage()
            output = compressor.output()
            self.assertEqual(expected, output)

        compare()

        filename = os.path.join(settings.COMPRESS_ROOT, "css", "one.css")
        test_css_content = "p { font-family: 'test' }"
        with open(filename, "a") as css:
            css.write("\n")
            css.write(test_css_content)
        compare()

        result_filename = os.path.join(
            settings.COMPRESS_ROOT, "CACHE", "css", "%s.css" % self.expected_css_hash
        )
        with open(result_filename, "r") as f:
            result = f.read()
        self.assertTrue(test_css_content not in result)



from django.apps import apps
from django.core.exceptions import ImproperlyConfigured

from compressor.conf import settings


if apps.is_installed("django.contrib.staticfiles"):
    from django.contrib.staticfiles import finders  # noqa

    if "compressor.finders.CompressorFinder" not in settings.STATICFILES_FINDERS:
        raise ImproperlyConfigured(
            "When using Django Compressor together with staticfiles, "
            "please add 'compressor.finders.CompressorFinder' to the "
            "STATICFILES_FINDERS setting."
        )
else:
    finders = None  # noqa



import io

from django.core.management import call_command
from django.core.management.base import CommandError
from django.test import TestCase


class TestMtimeCacheCommand(TestCase):
    # FIXME: add actual tests, improve the existing ones.

    exclusion_patterns = [
        "*CACHE*",
        "*custom*",
        "*066cd253eada.js",
        "*d728fc7f9301.js",
        "*8a0fed36c317.js",
        "test.txt*",
    ]

    def default_ignore(self):
        return ["--ignore=%s" % pattern for pattern in self.exclusion_patterns]

    def test_handle_no_args(self):
        with self.assertRaises(CommandError):
            call_command("mtime_cache")

    def test_handle_add(self):
        out = io.StringIO()
        with self.settings(CACHES={}):
            call_command("mtime_cache", "--add", *self.default_ignore(), stdout=out)
        output = out.getvalue()
        self.assertIn("Deleted mtimes of 20 files from the cache.", output)
        self.assertIn("Added mtimes of 20 files to cache.", output)

    def test_handle_clean(self):
        out = io.StringIO()
        with self.settings(CACHES={}):
            call_command("mtime_cache", "--clean", *self.default_ignore(), stdout=out)
        output = out.getvalue()
        self.assertIn("Deleted mtimes of 20 files from the cache.", output)
        self.assertNotIn("Added mtimes of 20 files to cache.", output)



import io
import os
import sys
from collections import defaultdict
from unittest import mock

from django.conf import settings
from django.test import override_settings, TestCase
from django.utils.encoding import smart_str

from compressor.cache import cache, get_hashed_content, get_hashed_mtime
from compressor.css import CssCompressor
from compressor.filters import CachedCompilerFilter, CompilerFilter
from compressor.filters.cleancss import CleanCSSFilter
from compressor.filters.closure import ClosureCompilerFilter
from compressor.filters.css_default import CssAbsoluteFilter, CssRelativeFilter
from compressor.filters.cssmin import CSSCompressorFilter, rCSSMinFilter
from compressor.filters.jsmin import CalmjsFilter, rJSMinFilter
from compressor.filters.template import TemplateFilter
from compressor.filters.yuglify import YUglifyCSSFilter, YUglifyJSFilter
from compressor.filters.yui import YUICSSFilter, YUIJSFilter
from compressor.tests.test_base import test_dir


def blankdict(*args, **kwargs):
    return defaultdict(lambda: "", *args, **kwargs)


@override_settings(COMPRESS_CACHEABLE_PRECOMPILERS=("text/css",))
class PrecompilerTestCase(TestCase):
    CHARSET = "utf-8"

    def setUp(self):
        self.test_precompiler = os.path.join(test_dir, "precompiler.py")
        self.setup_infile()
        self.cached_precompiler_args = dict(
            content=self.content,
            charset=self.CHARSET,
            filename=self.filename,
            mimetype="text/css",
        )

    def setup_infile(self, filename="static/css/one.css"):
        self.filename = os.path.join(test_dir, filename)
        with io.open(self.filename, encoding=self.CHARSET) as file:
            self.content = file.read()

    def test_precompiler_dict_options(self):
        command = "%s %s {option}" % (sys.executable, self.test_precompiler)
        option = (
            "option",
            "option",
        )
        CompilerFilter.options = dict([option])
        compiler = CompilerFilter(
            content=self.content,
            filename=self.filename,
            charset=self.CHARSET,
            command=command,
        )
        self.assertIn(option, compiler.options)

    def test_precompiler_infile_outfile(self):
        command = "%s %s -f {infile} -o {outfile}" % (
            sys.executable,
            self.test_precompiler,
        )
        compiler = CompilerFilter(
            content=self.content,
            filename=self.filename,
            charset=self.CHARSET,
            command=command,
        )
        self.assertEqual("body { color:#990; }", compiler.input())

    def test_precompiler_infile_with_spaces(self):
        self.setup_infile("static/css/filename with spaces.css")
        command = "%s %s -f {infile} -o {outfile}" % (
            sys.executable,
            self.test_precompiler,
        )
        compiler = CompilerFilter(
            content=self.content,
            filename=self.filename,
            charset=self.CHARSET,
            command=command,
        )
        self.assertEqual("body { color:#424242; }", compiler.input())

    def test_precompiler_infile_stdout(self):
        command = "%s %s -f {infile}" % (sys.executable, self.test_precompiler)
        compiler = CompilerFilter(
            content=self.content, filename=None, charset=None, command=command
        )
        self.assertEqual("body { color:#990; }%s" % os.linesep, compiler.input())

    def test_precompiler_stdin_outfile(self):
        command = "%s %s -o {outfile}" % (sys.executable, self.test_precompiler)
        compiler = CompilerFilter(
            content=self.content, filename=None, charset=None, command=command
        )
        self.assertEqual("body { color:#990; }", compiler.input())

    def test_precompiler_stdin_stdout(self):
        command = "%s %s" % (sys.executable, self.test_precompiler)
        compiler = CompilerFilter(
            content=self.content, filename=None, charset=None, command=command
        )
        self.assertEqual("body { color:#990; }%s" % os.linesep, compiler.input())

    def test_precompiler_stdin_stdout_filename(self):
        command = "%s %s" % (sys.executable, self.test_precompiler)
        compiler = CompilerFilter(
            content=self.content,
            filename=self.filename,
            charset=self.CHARSET,
            command=command,
        )
        self.assertEqual("body { color:#990; }%s" % os.linesep, compiler.input())

    def test_precompiler_output_unicode(self):
        command = "%s %s" % (sys.executable, self.test_precompiler)
        compiler = CompilerFilter(
            content=self.content, filename=self.filename, command=command
        )
        self.assertEqual(type(compiler.input()), str)

    def test_precompiler_cache(self):
        # The cache may already have data in it depending on the order the tests are
        # run, so start by clearing it:
        cache.clear()
        command = "%s %s -f {infile} -o {outfile}" % (
            sys.executable,
            self.test_precompiler,
        )
        compiler = CachedCompilerFilter(command=command, **self.cached_precompiler_args)
        self.assertEqual("body { color:#990; }", compiler.input())
        # We tell whether the precompiler actually ran by inspecting compiler.infile. If not None, the compiler had to
        # write the input out to the file for the external command. If None, it was in the cache and thus skipped.
        self.assertIsNotNone(compiler.infile)  # Not cached

        compiler = CachedCompilerFilter(command=command, **self.cached_precompiler_args)
        self.assertEqual("body { color:#990; }", compiler.input())
        self.assertIsNone(compiler.infile)  # Cached

        self.cached_precompiler_args[
            "content"
        ] += " "  # Invalidate cache by slightly changing content
        compiler = CachedCompilerFilter(command=command, **self.cached_precompiler_args)
        self.assertEqual("body { color:#990; }", compiler.input())
        self.assertIsNotNone(compiler.infile)  # Not cached

    @mock.patch("django.core.cache.backends.locmem.LocMemCache.get")
    def test_precompiler_cache_issue750(self, mock_cache):
        # emulate memcached and return string
        mock_cache.side_effect = lambda key: str("body { color:#990; }")
        command = "%s %s -f {infile} -o {outfile}" % (
            sys.executable,
            self.test_precompiler,
        )
        compiler = CachedCompilerFilter(command=command, **self.cached_precompiler_args)
        self.assertEqual("body { color:#990; }", compiler.input())
        self.assertEqual(
            type(compiler.input()), type(smart_str("body { color:#990; }"))
        )

    def test_precompiler_not_cacheable(self):
        command = "%s %s -f {infile} -o {outfile}" % (
            sys.executable,
            self.test_precompiler,
        )
        self.cached_precompiler_args["mimetype"] = "text/different"
        compiler = CachedCompilerFilter(command=command, **self.cached_precompiler_args)
        self.assertEqual("body { color:#990; }", compiler.input())
        self.assertIsNotNone(compiler.infile)  # Not cached

        compiler = CachedCompilerFilter(command=command, **self.cached_precompiler_args)
        self.assertEqual("body { color:#990; }", compiler.input())
        self.assertIsNotNone(compiler.infile)  # Not cached

    def test_precompiler_caches_empty_files(self):
        command = "%s %s -f {infile} -o {outfile}" % (
            sys.executable,
            self.test_precompiler,
        )
        compiler = CachedCompilerFilter(command=command, **self.cached_precompiler_args)
        self.assertEqual("body { color:#990; }", compiler.input())

        cache.set(compiler.get_cache_key(), "")
        compiler = CachedCompilerFilter(command=command, **self.cached_precompiler_args)
        self.assertEqual("", compiler.input())


class CSSCompressorTestCase(TestCase):
    def test_csscompressor_filter(self):
        content = """/*!
 * django-compressor
 * Copyright (c) 2009-2014 Django Compressor authors
 */
        p {


        background: rgb(51,102,153) url('../../images/image.gif');


        }
        """
        output = """/*!
 * django-compressor
 * Copyright (c) 2009-2014 Django Compressor authors
 */p{background:#369 url('../../images/image.gif')}"""
        self.assertEqual(output, CSSCompressorFilter(content).output())


class rCssMinTestCase(TestCase):
    def test_rcssmin_filter(self):
        content = """/*!
 * django-compressor
 * Copyright (c) 2009-2014 Django Compressor authors
 */
        p {


        background: rgb(51,102,153) url('../../images/image.gif');


        }
        """
        output = """/*!
 * django-compressor
 * Copyright (c) 2009-2014 Django Compressor authors
 */p{background:rgb(51,102,153) url('../../images/image.gif')}"""
        self.assertEqual(output, rCSSMinFilter(content).output())


class JsMinTestCase(TestCase):
    def test_jsmin_filter(self):
        content = """/*!
 * django-compressor
 * Copyright (c) 2009-2014 Django Compressor authors
 */
        var foo = "bar";"""
        output = """/*!
 * django-compressor
 * Copyright (c) 2009-2014 Django Compressor authors
 */var foo="bar";"""
        self.assertEqual(output, rJSMinFilter(content).output())


class CalmjsTestCase(TestCase):
    def test_calmjs_filter(self):
        content = """
        var foo = "bar";"""
        output = """var foo="bar";"""
        self.assertEqual(output, CalmjsFilter(content).output())


@override_settings(
    COMPRESS_ENABLED=True,
    COMPRESS_URL="/static/",
)
class CssAbsolutizingTestCase(TestCase):
    hashing_method = "mtime"
    hashing_func = staticmethod(get_hashed_mtime)
    template = (
        "p { background: url('%(url)simg/python.png%(query)s%(hash)s%(frag)s') }"
        "p { filter: Alpha(src='%(url)simg/python.png%(query)s%(hash)s%(frag)s') }"
    )
    filter_class = CssAbsoluteFilter

    @property
    def expected_url_prefix(self):
        return settings.COMPRESS_URL

    def setUp(self):
        self.override_settings = self.settings(
            COMPRESS_CSS_HASHING_METHOD=self.hashing_method
        )
        self.override_settings.__enter__()

    def tearDown(self):
        self.override_settings.__exit__(None, None, None)

    @override_settings(COMPRESS_CSS_HASHING_METHOD=None)
    def test_css_no_hash(self):
        filename = os.path.join(settings.COMPRESS_ROOT, "css/url/test.css")
        content = self.template % blankdict(url="../../")
        params = blankdict(
            {
                "url": self.expected_url_prefix,
            }
        )
        output = self.template % params
        filter = self.filter_class(content)
        self.assertEqual(
            output, filter.input(filename=filename, basename="css/url/test.css")
        )

    def test_css_absolute_filter(self):
        filename = os.path.join(settings.COMPRESS_ROOT, "css/url/test.css")
        imagefilename = os.path.join(settings.COMPRESS_ROOT, "img/python.png")
        content = self.template % blankdict(url="../../")
        params = blankdict(
            {
                "url": self.expected_url_prefix,
                "hash": "?" + self.hashing_func(imagefilename),
            }
        )
        output = self.template % params
        filter = self.filter_class(content)
        self.assertEqual(
            output, filter.input(filename=filename, basename="css/url/test.css")
        )

    def test_css_absolute_filter_url_fragment(self):
        filename = os.path.join(settings.COMPRESS_ROOT, "css/url/test.css")
        imagefilename = os.path.join(settings.COMPRESS_ROOT, "img/python.png")
        content = self.template % blankdict(url="../../", frag="#foo")
        params = blankdict(
            {
                "url": self.expected_url_prefix,
                "hash": "?" + self.hashing_func(imagefilename),
                "frag": "#foo",
            }
        )
        output = self.template % params
        filter = self.filter_class(content)
        self.assertEqual(
            output, filter.input(filename=filename, basename="css/url/test.css")
        )

    def test_css_absolute_filter_only_url_fragment(self):
        filename = os.path.join(settings.COMPRESS_ROOT, "css/url/test.css")
        content = "p { background: url('#foo') }"
        filter = self.filter_class(content)
        self.assertEqual(
            content, filter.input(filename=filename, basename="css/url/test.css")
        )

    def test_css_absolute_filter_only_url_fragment_wrap_double_quotes(self):
        filename = os.path.join(settings.COMPRESS_ROOT, "css/url/test.css")
        content = 'p { background: url("#foo") }'
        filter = self.filter_class(content)
        self.assertEqual(
            content, filter.input(filename=filename, basename="css/url/test.css")
        )

    def test_css_absolute_filter_querystring(self):
        filename = os.path.join(settings.COMPRESS_ROOT, "css/url/test.css")
        imagefilename = os.path.join(settings.COMPRESS_ROOT, "img/python.png")
        content = self.template % blankdict(url="../../", query="?foo")
        params = blankdict(
            {
                "url": self.expected_url_prefix,
                "query": "?foo",
                "hash": "&" + self.hashing_func(imagefilename),
            }
        )
        output = self.template % params
        filter = self.filter_class(content)
        self.assertEqual(
            output, filter.input(filename=filename, basename="css/url/test.css")
        )

    def test_css_absolute_filter_https(self):
        with self.settings(COMPRESS_URL="https://static.example.com/"):
            self.test_css_absolute_filter()

    def test_css_absolute_filter_relative_path(self):
        filename = os.path.join(
            settings.TEST_DIR,
            "whatever",
            "..",
            "static",
            "whatever/../css/url/test.css",
        )
        imagefilename = os.path.join(settings.COMPRESS_ROOT, "img/python.png")
        content = self.template % blankdict(url="../../")
        params = blankdict(
            {
                "url": self.expected_url_prefix,
                "hash": "?" + self.hashing_func(imagefilename),
            }
        )
        output = self.template % params
        filter = self.filter_class(content)
        self.assertEqual(
            output, filter.input(filename=filename, basename="css/url/test.css")
        )

    def test_css_absolute_filter_filename_outside_compress_root(self):
        filename = "/foo/bar/baz/test.css"
        content = self.template % blankdict(url="../qux/")
        params = blankdict(
            {
                "url": self.expected_url_prefix + "bar/qux/",
            }
        )
        output = self.template % params
        filter = self.filter_class(content)
        self.assertEqual(
            output, filter.input(filename=filename, basename="bar/baz/test.css")
        )

    def test_css_hunks(self):
        hash_python_png = self.hashing_func(
            os.path.join(settings.COMPRESS_ROOT, "img/python.png")
        )
        hash_add_png = self.hashing_func(
            os.path.join(settings.COMPRESS_ROOT, "img/add.png")
        )

        css1 = """\
p { background: url('%(compress_url)simg/python.png?%(hash)s'); }
p { background: url(%(compress_url)simg/python.png?%(hash)s); }
p { background: url(%(compress_url)simg/python.png?%(hash)s); }
p { background: url('%(compress_url)simg/python.png?%(hash)s'); }
p { filter: progid:DXImageTransform.Microsoft.AlphaImageLoader(src='%(compress_url)simg/python.png?%(hash)s'); }
""" % dict(
            compress_url=self.expected_url_prefix, hash=hash_python_png
        )

        css2 = """\
p { background: url('%(compress_url)simg/add.png?%(hash)s'); }
p { background: url(%(compress_url)simg/add.png?%(hash)s); }
p { background: url(%(compress_url)simg/add.png?%(hash)s); }
p { background: url('%(compress_url)simg/add.png?%(hash)s'); }
p { filter: progid:DXImageTransform.Microsoft.AlphaImageLoader(src='%(compress_url)simg/add.png?%(hash)s'); }
""" % dict(
            compress_url=self.expected_url_prefix, hash=hash_add_png
        )

        css = """
        <link rel="stylesheet" href="/static/css/url/url1.css" type="text/css">
        <link rel="stylesheet" href="/static/css/url/2/url2.css" type="text/css">
        """
        css_node = CssCompressor("css", css)

        self.assertEqual([css1, css2], list(css_node.hunks()))

    def test_guess_filename(self):
        url = "%s/img/python.png" % settings.COMPRESS_URL.rstrip("/")
        path = os.path.join(settings.COMPRESS_ROOT, "img/python.png")
        content = "p { background: url('%s') }" % url
        filter = self.filter_class(content)
        self.assertEqual(path, filter.guess_filename(url))

    def test_filenames_with_space(self):
        filename = os.path.join(settings.COMPRESS_ROOT, "css/url/test.css")
        imagefilename = os.path.join(settings.COMPRESS_ROOT, "img/add with spaces.png")

        template = "p { background: url('%(url)simg/add with spaces.png%(query)s%(hash)s%(frag)s') }"

        content = template % blankdict(url="../../")
        params = blankdict(
            {
                "url": self.expected_url_prefix,
                "hash": "?" + self.hashing_func(imagefilename),
            }
        )
        output = template % params
        filter = self.filter_class(content)
        self.assertEqual(
            output, filter.input(filename=filename, basename="css/url/test.css")
        )

    def test_does_not_change_nested_urls(self):
        css = """body { background-image: url("data:image/svg+xml;utf8,<svg><rect fill='url(%23gradient)'/></svg>");}"""
        filter = self.filter_class(css, filename="doesntmatter")
        self.assertEqual(
            css, filter.input(filename="doesntmatter", basename="doesntmatter")
        )

    def test_does_not_change_quotes_in_src(self):
        filename = os.path.join(settings.COMPRESS_ROOT, "css/url/test.css")
        hash_add_png = self.hashing_func(
            os.path.join(settings.COMPRESS_ROOT, "img/add.png")
        )
        css = """p { filter: Alpha(src="/img/add.png%(hash)s") }"""
        filter = self.filter_class(css % dict(hash=""))
        expected = css % dict(hash="?" + hash_add_png)
        self.assertEqual(
            expected, filter.input(filename=filename, basename="css/url/test.css")
        )


@override_settings(COMPRESS_URL="http://static.example.com/")
class CssAbsolutizingTestCaseWithDifferentURL(CssAbsolutizingTestCase):
    pass


class CssAbsolutizingTestCaseWithHash(CssAbsolutizingTestCase):
    hashing_method = "content"
    hashing_func = staticmethod(get_hashed_content)


@override_settings(
    COMPRESS_ENABLED=True,
    COMPRESS_URL="/static/",
    COMPRESS_FILTERS={"css": ["compressor.filters.css_default.CssRelativeFilter"]},
)
class CssRelativizingTestCase(CssAbsolutizingTestCase):
    filter_class = CssRelativeFilter
    expected_url_prefix = "../../"

    @override_settings(
        COMPRESS_CSS_HASHING_METHOD=None, COMPRESS_OUTPUT_DIR="CACHE/in/depth"
    )
    def test_nested_cache_dir(self):
        filename = os.path.join(settings.COMPRESS_ROOT, "css/url/test.css")
        content = self.template % blankdict(url="../../")
        params = blankdict(
            {
                "url": "../../../../",
            }
        )
        output = self.template % params
        filter = self.filter_class(content)
        self.assertEqual(
            output, filter.input(filename=filename, basename="css/url/test.css")
        )


@override_settings(
    COMPRESS_ENABLED=True,
    COMPRESS_FILTERS={
        "css": [
            "compressor.filters.css_default.CssAbsoluteFilter",
            "compressor.filters.datauri.CssDataUriFilter",
        ]
    },
    COMPRESS_URL="/static/",
    COMPRESS_CSS_HASHING_METHOD="mtime",
)
class CssDataUriTestCase(TestCase):
    def setUp(self):
        self.css = """
        <link rel="stylesheet" href="/static/css/datauri.css" type="text/css">
        """
        self.css_node = CssCompressor("css", self.css)

    def test_data_uris(self):
        datauri_hash = get_hashed_mtime(
            os.path.join(settings.COMPRESS_ROOT, "img/python.png")
        )
        out = [
            """.add { background-image: url("data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAAABGdBTUEAAK/INwWK6QAAABl0RVh0U29mdHdhcmUAQWRvYmUgSW1hZ2VSZWFkeXHJZTwAAAJvSURBVDjLpZPrS5NhGIf9W7YvBYOkhlkoqCklWChv2WyKik7blnNris72bi6dus0DLZ0TDxW1odtopDs4D8MDZuLU0kXq61CijSIIasOvv94VTUfLiB74fXngup7nvrnvJABJ/5PfLnTTdcwOj4RsdYmo5glBWP6iOtzwvIKSWstI0Wgx80SBblpKtE9KQs/We7EaWoT/8wbWP61gMmCH0lMDvokT4j25TiQU/ITFkek9Ow6+7WH2gwsmahCPdwyw75uw9HEO2gUZSkfyI9zBPCJOoJ2SMmg46N61YO/rNoa39Xi41oFuXysMfh36/Fp0b7bAfWAH6RGi0HglWNCbzYgJaFjRv6zGuy+b9It96N3SQvNKiV9HvSaDfFEIxXItnPs23BzJQd6DDEVM0OKsoVwBG/1VMzpXVWhbkUM2K4oJBDYuGmbKIJ0qxsAbHfRLzbjcnUbFBIpx/qH3vQv9b3U03IQ/HfFkERTzfFj8w8jSpR7GBE123uFEYAzaDRIqX/2JAtJbDat/COkd7CNBva2cMvq0MGxp0PRSCPF8BXjWG3FgNHc9XPT71Ojy3sMFdfJRCeKxEsVtKwFHwALZfCUk3tIfNR8XiJwc1LmL4dg141JPKtj3WUdNFJqLGFVPC4OkR4BxajTWsChY64wmCnMxsWPCHcutKBxMVp5mxA1S+aMComToaqTRUQknLTH62kHOVEE+VQnjahscNCy0cMBWsSI0TCQcZc5ALkEYckL5A5noWSBhfm2AecMAjbcRWV0pUTh0HE64TNf0mczcnnQyu/MilaFJCae1nw2fbz1DnVOxyGTlKeZft/Ff8x1BRssfACjTwQAAAABJRU5ErkJggg=="); }
.add-with-hash { background-image: url("data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAAABGdBTUEAAK/INwWK6QAAABl0RVh0U29mdHdhcmUAQWRvYmUgSW1hZ2VSZWFkeXHJZTwAAAJvSURBVDjLpZPrS5NhGIf9W7YvBYOkhlkoqCklWChv2WyKik7blnNris72bi6dus0DLZ0TDxW1odtopDs4D8MDZuLU0kXq61CijSIIasOvv94VTUfLiB74fXngup7nvrnvJABJ/5PfLnTTdcwOj4RsdYmo5glBWP6iOtzwvIKSWstI0Wgx80SBblpKtE9KQs/We7EaWoT/8wbWP61gMmCH0lMDvokT4j25TiQU/ITFkek9Ow6+7WH2gwsmahCPdwyw75uw9HEO2gUZSkfyI9zBPCJOoJ2SMmg46N61YO/rNoa39Xi41oFuXysMfh36/Fp0b7bAfWAH6RGi0HglWNCbzYgJaFjRv6zGuy+b9It96N3SQvNKiV9HvSaDfFEIxXItnPs23BzJQd6DDEVM0OKsoVwBG/1VMzpXVWhbkUM2K4oJBDYuGmbKIJ0qxsAbHfRLzbjcnUbFBIpx/qH3vQv9b3U03IQ/HfFkERTzfFj8w8jSpR7GBE123uFEYAzaDRIqX/2JAtJbDat/COkd7CNBva2cMvq0MGxp0PRSCPF8BXjWG3FgNHc9XPT71Ojy3sMFdfJRCeKxEsVtKwFHwALZfCUk3tIfNR8XiJwc1LmL4dg141JPKtj3WUdNFJqLGFVPC4OkR4BxajTWsChY64wmCnMxsWPCHcutKBxMVp5mxA1S+aMComToaqTRUQknLTH62kHOVEE+VQnjahscNCy0cMBWsSI0TCQcZc5ALkEYckL5A5noWSBhfm2AecMAjbcRWV0pUTh0HE64TNf0mczcnnQyu/MilaFJCae1nw2fbz1DnVOxyGTlKeZft/Ff8x1BRssfACjTwQAAAABJRU5ErkJggg=="); }
.python { background-image: url("/static/img/python.png?%s"); }
.datauri { background-image: url("data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAoAAAAKCAYAAACNMs+9AAAABGdBTUEAALGPC/xhBQAAAAlwSFlzAAALEwAACxMBAJqcGAAAAAd0SU1FB9YGARc5KB0XV+IAAAAddEVYdENvbW1lbnQAQ3JlYXRlZCB3aXRoIFRoZSBHSU1Q72QlbgAAAF1JREFUGNO9zL0NglAAxPEfdLTs4BZM4DIO4C7OwQg2JoQ9LE1exdlYvBBeZ7jqch9//q1uH4TLzw4d6+ErXMMcXuHWxId3KOETnnXXV6MJpcq2MLaI97CER3N0 vr4MkhoXe0rZigAAAABJRU5ErkJggg=="); }
"""
            % datauri_hash
        ]
        self.assertEqual(out, list(self.css_node.hunks()))


class TemplateTestCase(TestCase):
    @override_settings(
        COMPRESS_TEMPLATE_FILTER_CONTEXT={"stuff": "thing", "gimmick": "bold"}
    )
    def test_template_filter(self):
        content = """
        #content {background-image: url("{{ STATIC_URL|default:stuff }}/images/bg.png");}
        #footer {font-weight: {{ gimmick }};}
        """
        input = """
        #content {background-image: url("thing/images/bg.png");}
        #footer {font-weight: bold;}
        """
        self.assertEqual(input, TemplateFilter(content).input())


class SpecializedFiltersTest(TestCase):
    """
    Test to check the Specializations of filters.
    """

    def test_closure_filter(self):
        filter = ClosureCompilerFilter("")
        self.assertEqual(
            filter.options,
            (("binary", str("java -jar compiler.jar")), ("args", str(""))),
        )

    def test_yuglify_filters(self):
        filter = YUglifyCSSFilter("")
        self.assertEqual(filter.command, "{binary} {args} --type=css")
        self.assertEqual(
            filter.options, (("binary", str("yuglify")), ("args", str("--terminal")))
        )

        filter = YUglifyJSFilter("")
        self.assertEqual(filter.command, "{binary} {args} --type=js")
        self.assertEqual(
            filter.options, (("binary", str("yuglify")), ("args", str("--terminal")))
        )

    def test_yui_filters(self):
        filter = YUICSSFilter("")
        self.assertEqual(filter.command, "{binary} {args} --type=css")
        self.assertEqual(
            filter.options,
            (("binary", str("java -jar yuicompressor.jar")), ("args", str(""))),
        )

        filter = YUIJSFilter("", verbose=1)
        self.assertEqual(filter.command, "{binary} {args} --type=js --verbose")
        self.assertEqual(
            filter.options,
            (
                ("binary", str("java -jar yuicompressor.jar")),
                ("args", str("")),
                ("verbose", 1),
            ),
        )

    def test_clean_css_filter(self):
        filter = CleanCSSFilter("")
        self.assertEqual(
            filter.options, (("binary", str("cleancss")), ("args", str("")))
        )



from compressor.conf import settings
from compressor.base import Compressor, SOURCE_HUNK, SOURCE_FILE


class JsCompressor(Compressor):

    output_mimetypes = {"text/javascript"}

    def split_contents(self):
        if self.split_content:
            return self.split_content
        self.extra_nodes = []
        for elem in self.parser.js_elems():
            attribs = self.parser.elem_attribs(elem)
            if "src" in attribs:
                basename = self.get_basename(attribs["src"])
                filename = self.get_filename(basename)
                content = (SOURCE_FILE, filename, basename, elem)
            else:
                content = (SOURCE_HUNK, self.parser.elem_content(elem), None, elem)
            self.split_content.append(content)
            if "async" in attribs:
                extra = " async"
            elif "defer" in attribs:
                extra = " defer"
            else:
                extra = ""
            # Append to the previous node if it had the same attribute
            append_to_previous = self.extra_nodes and self.extra_nodes[-1][0] == extra
            if append_to_previous and settings.COMPRESS_ENABLED:
                self.extra_nodes[-1][1].split_content.append(content)
            else:
                node = self.copy(content=self.parser.elem_str(elem))
                node.split_content.append(content)
                self.extra_nodes.append((extra, node))
        return self.split_content

    def output(self, *args, **kwargs):
        if (
            settings.COMPRESS_ENABLED
            or settings.COMPRESS_PRECOMPILERS
            or kwargs.get("forced", False)
        ):
            self.split_contents()
            if hasattr(self, "extra_nodes"):
                ret = []
                for extra, subnode in self.extra_nodes:
                    subnode.extra_context.update({"extra": extra})
                    ret.append(subnode.output(*args, **kwargs))
                return "\n".join(ret)
        return super().output(*args, **kwargs)

    def filter_input(self, forced=False):
        """
        Passes each hunk (file or code) to the 'input' methods
        of the compressor filters.
        """
        content = []
        for hunk in self.hunks(forced):
            # If a file ends with a function call, say, console.log()
            # but doesn't have a semicolon, and the next file starts with
            # a (, the individual files are ok, but when combined you get an
            # error like TypeError...
            # Forcing a semicolon in between fixes it.
            if settings.COMPRESS_ENABLED or forced:
                hunk += ";"
            content.append(hunk)
        return content



from django.test import TestCase
from django.test.utils import override_settings

from compressor.conf import settings
from compressor.tests.test_base import css_tag


class TestJinja2CompressorExtension(TestCase):
    """
    Test case for jinja2 extension.

    .. note::
       At tests we need to make some extra care about whitespace. Please note
       that we use jinja2 specific controls (*minus* character at block's
       beginning or end). For more information see jinja2 documentation.
    """

    def assertStrippedEqual(self, result, expected):
        self.assertEqual(
            result.strip(),
            expected.strip(),
            "%r != %r" % (result.strip(), expected.strip()),
        )

    def setUp(self):
        import jinja2

        self.jinja2 = jinja2
        from compressor.contrib.jinja2ext import CompressorExtension

        self.env = self.jinja2.Environment(extensions=[CompressorExtension])

    def test_error_raised_if_no_arguments_given(self):
        self.assertRaises(
            self.jinja2.exceptions.TemplateSyntaxError,
            self.env.from_string,
            "{% compress %}Foobar{% endcompress %}",
        )

    def test_error_raised_if_wrong_kind_given(self):
        self.assertRaises(
            self.jinja2.exceptions.TemplateSyntaxError,
            self.env.from_string,
            "{% compress foo %}Foobar{% endcompress %}",
        )

    def test_error_raised_if_wrong_closing_kind_given(self):
        self.assertRaises(
            self.jinja2.exceptions.TemplateSyntaxError,
            self.env.from_string,
            "{% compress js %}Foobar{% endcompress css %}",
        )

    def test_error_raised_if_wrong_mode_given(self):
        self.assertRaises(
            self.jinja2.exceptions.TemplateSyntaxError,
            self.env.from_string,
            "{% compress css foo %}Foobar{% endcompress %}",
        )

    @override_settings(COMPRESS_ENABLED=False)
    def test_compress_is_disabled(self):
        tag_body = "\n".join(
            [
                '<link rel="stylesheet" href="css/one.css" type="text/css" charset="utf-8">',
                '<style type="text/css">p { border:5px solid green;}</style>',
                '<link rel="stylesheet" href="css/two.css" type="text/css" charset="utf-8">',
            ]
        )
        template_string = "{% compress css %}" + tag_body + "{% endcompress %}"
        template = self.env.from_string(template_string)
        self.assertEqual(tag_body, template.render())

        # Test with explicit kind
        template_string = "{% compress css %}" + tag_body + "{% endcompress css %}"
        template = self.env.from_string(template_string)
        self.assertEqual(tag_body, template.render())

    def test_empty_tag(self):
        template = self.env.from_string(
            """{% compress js %}{% block js %}{% endblock %}{% endcompress %}"""
        )
        context = {"STATIC_URL": settings.COMPRESS_URL}
        self.assertEqual("", template.render(context))

    def test_empty_tag_with_kind(self):
        template = self.env.from_string(
            """{% compress js %}{% block js %}
        {% endblock %}{% endcompress js %}"""
        )
        context = {"STATIC_URL": settings.COMPRESS_URL}
        self.assertEqual("", template.render(context))

    def test_css_tag(self):
        template = self.env.from_string(
            """{% compress css -%}
        <link rel="stylesheet" href="{{ STATIC_URL }}css/one.css" type="text/css" charset="utf-8">
        <style type="text/css">p { border:5px solid green;}</style>
        <link rel="stylesheet" href="{{ STATIC_URL }}css/two.css" type="text/css" charset="utf-8">
        {% endcompress %}"""
        )
        context = {"STATIC_URL": settings.COMPRESS_URL}
        out = css_tag("/static/CACHE/css/output.600674ea1d3d.css")
        self.assertEqual(out, template.render(context))

    def test_nonascii_css_tag(self):
        template = self.env.from_string(
            """{% compress css -%}
        <link rel="stylesheet" href="{{ STATIC_URL }}css/nonasc.css" type="text/css" charset="utf-8">
        <style type="text/css">p { border:5px solid green;}</style>
        {% endcompress %}"""
        )
        context = {"STATIC_URL": settings.COMPRESS_URL}
        out = css_tag("/static/CACHE/css/output.d5444a1ab4a3.css")
        self.assertEqual(out, template.render(context))

    def test_js_tag(self):
        template = self.env.from_string(
            """{% compress js -%}
        <script src="{{ STATIC_URL }}js/one.js" type="text/javascript" charset="utf-8"></script>
        <script type="text/javascript" charset="utf-8">obj.value = "value";</script>
        {% endcompress %}"""
        )
        context = {"STATIC_URL": settings.COMPRESS_URL}
        out = '<script src="/static/CACHE/js/output.8a0fed36c317.js"></script>'
        self.assertEqual(out, template.render(context))

    def test_nonascii_js_tag(self):
        template = self.env.from_string(
            """{% compress js -%}
        <script src="{{ STATIC_URL }}js/nonasc.js" type="text/javascript" charset="utf-8"></script>
        <script type="text/javascript" charset="utf-8">var test_value = "\u2014";</script>
        {% endcompress %}"""
        )
        context = {"STATIC_URL": settings.COMPRESS_URL}
        out = '<script src="/static/CACHE/js/output.8c00f1cf1e0a.js"></script>'
        self.assertEqual(out, template.render(context))

    def test_nonascii_latin1_js_tag(self):
        template = self.env.from_string(
            """{% compress js -%}
        <script src="{{ STATIC_URL }}js/nonasc-latin1.js" type="text/javascript" charset="latin-1"></script>
        <script type="text/javascript">var test_value = "\u2014";</script>
        {% endcompress %}"""
        )
        context = {"STATIC_URL": settings.COMPRESS_URL}
        out = '<script src="/static/CACHE/js/output.06a98ccfd380.js"></script>'
        self.assertEqual(out, template.render(context))

    def test_css_inline(self):
        template = self.env.from_string(
            """{% compress css, inline -%}
        <link rel="stylesheet" href="{{ STATIC_URL }}css/one.css" type="text/css" charset="utf-8">
        <style type="text/css">p { border:5px solid green;}</style>
        {% endcompress %}"""
        )
        context = {"STATIC_URL": settings.COMPRESS_URL}
        out = (
            '<style type="text/css">body{background:#990}'
            "p{border:5px solid green}</style>"
        )
        self.assertEqual(out, template.render(context))

    def test_js_inline(self):
        template = self.env.from_string(
            """{% compress js, inline -%}
        <script src="{{ STATIC_URL }}js/one.js" type="text/css" type="text/javascript" charset="utf-8"></script>
        <script type="text/javascript" charset="utf-8">obj.value = "value";</script>
        {% endcompress %}"""
        )
        context = {"STATIC_URL": settings.COMPRESS_URL}
        out = '<script>obj={};;obj.value="value";;</script>'
        self.assertEqual(out, template.render(context))

    def test_nonascii_inline_css(self):
        with self.settings(COMPRESS_ENABLED=False):
            template = self.env.from_string(
                "{% compress css %}"
                '<style type="text/css">'
                "/* русский текст */"
                "</style>{% endcompress %}"
            )
        out = '<link rel="stylesheet" href="/static/CACHE/css/output.e3b0c44298fc.css" type="text/css">'
        context = {"STATIC_URL": settings.COMPRESS_URL}
        self.assertEqual(out, template.render(context))



from django.core.exceptions import ImproperlyConfigured
from django.utils.encoding import smart_str
from django.utils.functional import cached_property

from compressor.exceptions import ParserError
from compressor.parser import ParserBase


class Html5LibParser(ParserBase):
    def __init__(self, content):
        super().__init__(content)
        import html5lib

        self.html5lib = html5lib

    def _serialize(self, elem):
        return self.html5lib.serialize(
            elem,
            tree="etree",
            quote_attr_values="always",
            omit_optional_tags=False,
        )

    def _find(self, *names):
        for elem in self.html:
            if elem.tag in names:
                yield elem

    @cached_property
    def html(self):
        try:
            return self.html5lib.parseFragment(self.content, treebuilder="etree")
        except ImportError as err:
            raise ImproperlyConfigured("Error while importing html5lib: %s" % err)
        except Exception as err:
            raise ParserError("Error while initializing Parser: %s" % err)

    def css_elems(self):
        return self._find(
            "{http://www.w3.org/1999/xhtml}link", "{http://www.w3.org/1999/xhtml}style"
        )

    def js_elems(self):
        return self._find("{http://www.w3.org/1999/xhtml}script")

    def elem_attribs(self, elem):
        return elem.attrib

    def elem_content(self, elem):
        return smart_str(elem.text)

    def elem_name(self, elem):
        if "}" in elem.tag:
            return elem.tag.split("}")[1]
        return elem.tag

    def elem_str(self, elem):
        # This method serializes HTML in a way that does not pass all tests.
        # However, this method is only called in tests anyway, so it doesn't
        # really matter.
        return smart_str(self._serialize(elem))



import json
import hashlib
import os
import socket
import time
from importlib import import_module

from django.core.cache import caches
from django.core.files.base import ContentFile
from django.utils.encoding import force_str, smart_bytes
from django.utils.functional import SimpleLazyObject

from compressor.conf import settings
from compressor.storage import default_offline_manifest_storage
from compressor.utils import get_mod_func

_cachekey_func = None


def get_hexdigest(plaintext, length=None):
    digest = hashlib.sha256(smart_bytes(plaintext)).hexdigest()
    if length:
        return digest[:length]
    return digest


def simple_cachekey(key):
    return "django_compressor.%s" % force_str(key)


def socket_cachekey(key):
    return "django_compressor.%s.%s" % (socket.gethostname(), force_str(key))


def get_cachekey(*args, **kwargs):
    global _cachekey_func
    if _cachekey_func is None:
        try:
            mod_name, func_name = get_mod_func(settings.COMPRESS_CACHE_KEY_FUNCTION)
            _cachekey_func = getattr(import_module(mod_name), func_name)
        except (AttributeError, ImportError, TypeError) as e:
            raise ImportError(
                "Couldn't import cache key function %s: %s"
                % (settings.COMPRESS_CACHE_KEY_FUNCTION, e)
            )
    return _cachekey_func(*args, **kwargs)


def get_mtime_cachekey(filename):
    return get_cachekey("mtime.%s" % get_hexdigest(filename))


def get_offline_hexdigest(render_template_string):
    return get_hexdigest(
        # Make the hexdigest determination independent of STATIC_URL
        render_template_string.replace(
            # Cast ``settings.STATIC_URL`` to a string to allow it to be
            # a string-alike object to e.g. add ``SCRIPT_NAME`` WSGI param
            # as a *path prefix* to the output URL.
            # See https://code.djangoproject.com/ticket/25598.
            str(settings.STATIC_URL),
            "",
        )
    )


def get_offline_cachekey(source):
    return get_cachekey("offline.%s" % get_offline_hexdigest(source))


_offline_manifest = None


def get_offline_manifest():
    global _offline_manifest
    if _offline_manifest is None:
        filename = settings.COMPRESS_OFFLINE_MANIFEST
        if default_offline_manifest_storage.exists(filename):
            with default_offline_manifest_storage.open(filename) as fp:
                _offline_manifest = json.loads(fp.read().decode("utf8"))
        else:
            _offline_manifest = {}
    return _offline_manifest


def flush_offline_manifest():
    global _offline_manifest
    _offline_manifest = None


def write_offline_manifest(manifest):
    content = json.dumps(manifest, indent=2).encode("utf8")
    default_offline_manifest_storage.save(
        settings.COMPRESS_OFFLINE_MANIFEST, ContentFile(content)
    )
    flush_offline_manifest()


def get_templatetag_cachekey(compressor, mode, kind):
    return get_cachekey("templatetag.%s.%s.%s" % (compressor.cachekey, mode, kind))


def get_mtime(filename):
    if settings.COMPRESS_MTIME_DELAY:
        key = get_mtime_cachekey(filename)
        mtime = cache.get(key)
        if mtime is None:
            mtime = os.path.getmtime(filename)
            cache.set(key, mtime, settings.COMPRESS_MTIME_DELAY)
        return mtime
    return os.path.getmtime(filename)


def get_hashed_mtime(filename, length=12):
    try:
        filename = os.path.realpath(filename)
        mtime = str(int(get_mtime(filename)))
    except OSError:
        return None
    return get_hexdigest(mtime, length)


def get_hashed_content(filename, length=12):
    try:
        filename = os.path.realpath(filename)
    except OSError:
        return None

    # should we make sure that file is utf-8 encoded?
    with open(filename, "rb") as file:
        return get_hexdigest(file.read(), length)


def get_precompiler_cachekey(command, contents):
    return hashlib.sha1(
        smart_bytes("precompiler.%s.%s" % (command, contents))
    ).hexdigest()


def cache_get(key):
    packed_val = cache.get(key)
    if packed_val is None:
        return None
    val, refresh_time, refreshed = packed_val
    if (time.time() > refresh_time) and not refreshed:
        # Store the stale value while the cache
        # revalidates for another MINT_DELAY seconds.
        cache_set(key, val, refreshed=True, timeout=settings.COMPRESS_MINT_DELAY)
        return None
    return val


def cache_set(key, val, refreshed=False, timeout=None):
    if timeout is None:
        timeout = settings.COMPRESS_REBUILD_TIMEOUT
    refresh_time = timeout + time.time()
    real_timeout = timeout + settings.COMPRESS_MINT_DELAY
    packed_val = (val, refresh_time, refreshed)
    return cache.set(key, packed_val, real_timeout)


cache = SimpleLazyObject(lambda: caches[settings.COMPRESS_CACHE_BACKEND])



from compressor.conf import settings
from compressor.filters import CompilerFilter


class ClosureCompilerFilter(CompilerFilter):
    command = "{binary} {args}"
    options = (
        ("binary", settings.COMPRESS_CLOSURE_COMPILER_BINARY),
        ("args", settings.COMPRESS_CLOSURE_COMPILER_ARGUMENTS),
    )



import os
import sys
from unittest.mock import Mock

from django.conf import settings
from django.template import Context, Template, TemplateSyntaxError
from django.test import override_settings, TestCase
from sekizai.context import SekizaiContext

from compressor.signals import post_compress
from compressor.tests.test_base import css_tag, test_dir


def render(template_string, context_dict=None, context=None):
    """
    A shortcut for testing template output.
    """
    if context_dict is None:
        context_dict = {}
    if context is None:
        context = Context
    c = context(context_dict)
    t = Template(template_string)
    return t.render(c).strip()


@override_settings(COMPRESS_ENABLED=True)
class TemplatetagTestCase(TestCase):
    def setUp(self):
        self.context = {"STATIC_URL": settings.COMPRESS_URL}

    def test_empty_tag(self):
        template = """{% load compress %}{% compress js %}{% block js %}
        {% endblock %}{% endcompress %}"""
        self.assertEqual("", render(template, self.context))

    def test_css_tag(self):
        template = """{% load compress %}{% compress css %}
<link rel="stylesheet" href="{{ STATIC_URL }}css/one.css" type="text/css">
<style type="text/css">p { border:5px solid green;}</style>
<link rel="stylesheet" href="{{ STATIC_URL }}css/two.css" type="text/css">
{% endcompress %}"""
        out = css_tag("/static/CACHE/css/output.600674ea1d3d.css")
        self.assertEqual(out, render(template, self.context))

    def test_css_tag_with_block(self):
        template = """{% load compress %}{% compress css file block_name %}
<link rel="stylesheet" href="{{ STATIC_URL }}css/one.css" type="text/css">
<style type="text/css">p { border:5px solid blue;}</style>
<link rel="stylesheet" href="{{ STATIC_URL }}css/two.css" type="text/css">
{% endcompress %}"""
        out = css_tag("/static/CACHE/css/block_name.a1e074d0c4ac.css")
        self.assertEqual(out, render(template, self.context))

    def test_missing_rel_leaves_empty_result(self):
        template = """{% load compress %}{% compress css %}
<link href="{{ STATIC_URL }}css/one.css" type="text/css">
{% endcompress %}"""
        out = ""
        self.assertEqual(out, render(template, self.context))

    def test_missing_rel_only_on_one_resource(self):
        template = """{% load compress %}{% compress css %}
<link href="{{ STATIC_URL }}css/wontmatter.css" type="text/css">
<link rel="stylesheet" href="{{ STATIC_URL }}css/one.css" type="text/css">
<style type="text/css">p { border:5px solid green;}</style>
<link rel="stylesheet" href="{{ STATIC_URL }}css/two.css" type="text/css">
{% endcompress %}"""
        out = css_tag("/static/CACHE/css/output.600674ea1d3d.css")
        self.assertEqual(out, render(template, self.context))

    def test_uppercase_rel(self):
        template = """{% load compress %}{% compress css %}
<link rel="StyleSheet" href="{{ STATIC_URL }}css/one.css" type="text/css">
<style type="text/css">p { border:5px solid green;}</style>
<link rel="StyleSheet" href="{{ STATIC_URL }}css/two.css" type="text/css">
{% endcompress %}"""
        out = css_tag("/static/CACHE/css/output.600674ea1d3d.css")
        self.assertEqual(out, render(template, self.context))

    def test_nonascii_css_tag(self):
        template = """{% load compress %}{% compress css %}
        <link rel="stylesheet" href="{{ STATIC_URL }}css/nonasc.css" type="text/css">
        <style type="text/css">p { border:5px solid green;}</style>
        {% endcompress %}
        """
        out = css_tag("/static/CACHE/css/output.d5444a1ab4a3.css")
        self.assertEqual(out, render(template, self.context))

    def test_js_tag(self):
        template = """{% load compress %}{% compress js %}
        <script src="{{ STATIC_URL }}js/one.js" type="text/javascript"></script>
        <script type="text/javascript">obj.value = "value";</script>
        {% endcompress %}
        """
        out = '<script src="/static/CACHE/js/output.8a0fed36c317.js"></script>'
        self.assertEqual(out, render(template, self.context))

    def test_nonascii_js_tag(self):
        template = """{% load compress %}{% compress js %}
        <script src="{{ STATIC_URL }}js/nonasc.js" type="text/javascript"></script>
        <script type="text/javascript">var test_value = "\u2014";</script>
        {% endcompress %}
        """
        out = '<script src="/static/CACHE/js/output.8c00f1cf1e0a.js"></script>'
        self.assertEqual(out, render(template, self.context))

    def test_nonascii_latin1_js_tag(self):
        template = """{% load compress %}{% compress js %}
        <script src="{{ STATIC_URL }}js/nonasc-latin1.js" type="text/javascript" charset="latin-1"></script>
        <script type="text/javascript">var test_value = "\u2014";</script>
        {% endcompress %}
        """
        out = '<script src="/static/CACHE/js/output.06a98ccfd380.js"></script>'
        self.assertEqual(out, render(template, self.context))

    def test_compress_tag_with_illegal_arguments(self):
        template = """{% load compress %}{% compress pony %}
        <script type="pony/application">unicorn</script>
        {% endcompress %}"""
        self.assertRaises(TemplateSyntaxError, render, template, {})

    @override_settings(COMPRESS_DEBUG_TOGGLE="togglecompress")
    def test_debug_toggle(self):
        template = """{% load compress %}{% compress js %}
        <script src="{{ STATIC_URL }}js/one.js" type="text/javascript"></script>
        <script type="text/javascript">obj.value = "value";</script>
        {% endcompress %}
        """

        class MockDebugRequest:
            GET = {settings.COMPRESS_DEBUG_TOGGLE: "true"}

        context = dict(self.context, request=MockDebugRequest())
        out = """<script src="/static/js/one.js" type="text/javascript"></script>
        <script type="text/javascript">obj.value = "value";</script>"""
        self.assertEqual(out, render(template, context))

    def test_inline(self):
        template = """{% load compress %}{% compress js inline %}
        <script src="{{ STATIC_URL }}js/one.js" type="text/javascript"></script>
        <script type="text/javascript">obj.value = "value";</script>
        {% endcompress %}{% compress css inline %}
        <link rel="stylesheet" href="{{ STATIC_URL }}css/one.css" type="text/css">
        <style type="text/css">p { border:5px solid green;}</style>
        <link rel="stylesheet" href="{{ STATIC_URL }}css/two.css" type="text/css">
        {% endcompress %}"""
        out_js = '<script>obj={};;obj.value="value";;</script>'
        out_css = (
            '<style type="text/css">body{background:#990}'
            "p{border:5px solid green}"
            "body{color:#fff}</style>"
        )
        self.assertEqual(out_js + out_css, render(template, self.context))

    def test_named_compress_tag(self):
        template = """{% load compress %}{% compress js inline foo %}
        <script type="text/javascript">obj.value = "value";</script>
        {% endcompress %}
        """

        def listener(sender, **kwargs):
            pass

        callback = Mock(wraps=listener)
        post_compress.connect(callback)
        render(template)
        args, kwargs = callback.call_args
        context = kwargs["context"]
        self.assertEqual("foo", context["compressed"]["name"])

    def test_sekizai_only_once(self):
        template = """{% load sekizai_tags %}{% addtoblock "js" %}
        <script type="text/javascript">var tmpl="{% templatetag openblock %} if x == 3 %}x IS 3{% templatetag openblock %} endif %}"</script>
        {% endaddtoblock %}{% render_block "js" postprocessor "compressor.contrib.sekizai.compress" %}
        """
        out = '<script src="/static/CACHE/js/output.ffc39dec05fd.js"></script>'
        self.assertEqual(out, render(template, self.context, SekizaiContext))


class PrecompilerTemplatetagTestCase(TestCase):
    def setUp(self):
        precompiler = os.path.join(test_dir, "precompiler.py")
        python = sys.executable

        override_settings = {
            "COMPRESS_ENABLED": True,
            "COMPRESS_PRECOMPILERS": (
                ("text/coffeescript", "%s %s" % (python, precompiler)),
                ("text/less", "%s %s" % (python, precompiler)),
            ),
        }
        self.override_settings = self.settings(**override_settings)
        self.override_settings.__enter__()

        self.context = {"STATIC_URL": settings.COMPRESS_URL}

    def tearDown(self):
        self.override_settings.__exit__(None, None, None)

    def test_compress_coffeescript_tag(self):
        template = """{% load compress %}{% compress js %}
            <script type="text/coffeescript"># this is a comment.</script>
            {% endcompress %}"""
        out = script(src="/static/CACHE/js/output.ec862f0ff42c.js")
        self.assertEqual(out, render(template, self.context))

    def test_compress_coffeescript_tag_and_javascript_tag(self):
        template = """{% load compress %}{% compress js %}
            <script type="text/coffeescript"># this is a comment.</script>
            <script type="text/javascript"># this too is a comment.</script>
            {% endcompress %}"""
        out = script(src="/static/CACHE/js/output.fb4a0d84e914.js")
        self.assertEqual(out, render(template, self.context))

    @override_settings(COMPRESS_ENABLED=False)
    def test_coffeescript_and_js_tag_with_compress_enabled_equals_false(self):
        template = """{% load compress %}{% compress js %}
            <script type="text/coffeescript"># this is a comment.</script>
            <script type="text/javascript"># this too is a comment.</script>
            {% endcompress %}"""
        out = (
            script("# this is a comment.\n")
            + "\n"
            + script("# this too is a comment.", scripttype="text/javascript")
        )
        self.assertEqual(out, render(template, self.context))

    @override_settings(COMPRESS_ENABLED=False)
    def test_compress_coffeescript_tag_compress_enabled_is_false(self):
        template = """{% load compress %}{% compress js %}
            <script type="text/coffeescript"># this is a comment.</script>
            {% endcompress %}"""
        out = script("# this is a comment.\n")
        self.assertEqual(out, render(template, self.context))

    @override_settings(COMPRESS_ENABLED=False)
    def test_compress_coffeescript_file_tag_compress_enabled_is_false(self):
        template = """
        {% load compress %}{% compress js %}
        <script type="text/coffeescript" src="{{ STATIC_URL }}js/one.coffee">
        </script>
        {% endcompress %}"""

        out = script(src="/static/CACHE/js/one.4b3570601b8c.js")
        self.assertEqual(out, render(template, self.context))

    @override_settings(COMPRESS_ENABLED=False)
    def test_multiple_file_order_conserved(self):
        template = """
        {% load compress %}{% compress js %}
        <script type="text/coffeescript" src="{{ STATIC_URL }}js/one.coffee">
        </script>
        <script src="{{ STATIC_URL }}js/one.js"></script>
        <script type="text/coffeescript" src="{{ STATIC_URL }}js/one.js">
        </script>
        {% endcompress %}"""

        out = "\n".join(
            [
                script(src="/static/CACHE/js/one.4b3570601b8c.js"),
                script(scripttype="", src="/static/js/one.js"),
                script(src="/static/CACHE/js/one.8ab93aace8fa.js"),
            ]
        )

        self.assertEqual(out, render(template, self.context))

    @override_settings(COMPRESS_ENABLED=False)
    def test_css_multiple_files_disabled_compression(self):
        assert settings.COMPRESS_PRECOMPILERS
        template = """
        {% load compress %}{% compress css %}
        <link rel="stylesheet" type="text/css" href="{{ STATIC_URL }}css/one.css"></link>
        <link rel="stylesheet" type="text/css" href="{{ STATIC_URL }}css/two.css"></link>
        {% endcompress %}"""

        out = "".join(
            [
                '<link rel="stylesheet" type="text/css" href="/static/css/one.css">',
                '<link rel="stylesheet" type="text/css" href="/static/css/two.css">',
            ]
        )

        self.assertEqual(out, render(template, self.context))

    @override_settings(COMPRESS_ENABLED=False)
    def test_css_multiple_files_mixed_precompile_disabled_compression(self):
        assert settings.COMPRESS_PRECOMPILERS
        template = """
        {% load compress %}{% compress css %}
        <link rel="stylesheet" type="text/css" href="{{ STATIC_URL }}css/one.css"/>
        <link rel="stylesheet" type="text/css" href="{{ STATIC_URL }}css/two.css"/>
        <link rel="stylesheet" type="text/less" href="{{ STATIC_URL }}css/url/test.css"/>
        {% endcompress %}"""

        out = "".join(
            [
                '<link rel="stylesheet" type="text/css" href="/static/css/one.css">',
                '<link rel="stylesheet" type="text/css" href="/static/css/two.css">',
                '<link rel="stylesheet" href="/static/CACHE/css/test.222f958fb191.css" type="text/css">',
            ]
        )
        self.assertEqual(out, render(template, self.context))


def script(content="", src="", scripttype=""):
    """
    returns a unicode text html script element.

    >>> script('#this is a comment', scripttype="text/applescript")
    '<script type="text/applescript">#this is a comment</script>'
    """
    out_script = "<script "
    if scripttype:
        out_script += 'type="%s" ' % scripttype
    if src:
        out_script += 'src="%s" ' % src
    return out_script[:-1] + ">%s</script>" % content



import os
import re
import posixpath

from compressor.cache import get_hashed_mtime, get_hashed_content
from compressor.conf import settings
from compressor.filters import FilterBase, FilterError

URL_PATTERN = re.compile(
    r"""
    url\(
    \s*      # any amount of whitespace
    ([\'"]?) # optional quote
    (.*?)    # any amount of anything, non-greedily (this is the actual url)
    \1       # matching quote (or nothing if there was none)
    \s*      # any amount of whitespace
    \)""",
    re.VERBOSE,
)
SRC_PATTERN = re.compile(r'src=([\'"])(.*?)\1')
SCHEMES = ("http://", "https://", "/")


class CssAbsoluteFilter(FilterBase):

    run_with_compression_disabled = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.root = settings.COMPRESS_ROOT
        self.url = settings.COMPRESS_URL.rstrip("/")
        self.url_path = self.url
        self.has_scheme = False

    def input(self, filename=None, basename=None, **kwargs):
        if not filename:
            return self.content
        self.path = basename.replace(os.sep, "/")
        self.path = self.path.lstrip("/")
        if self.url.startswith(("http://", "https://")):
            self.has_scheme = True
            parts = self.url.split("/")
            self.url = "/".join(parts[2:])
            self.url_path = "/%s" % "/".join(parts[3:])
            self.protocol = "%s/" % "/".join(parts[:2])
            self.host = parts[2]
        self.directory_name = "/".join((self.url, os.path.dirname(self.path)))
        return SRC_PATTERN.sub(
            self.src_converter, URL_PATTERN.sub(self.url_converter, self.content)
        )

    def guess_filename(self, url):
        local_path = url
        if self.has_scheme:
            # COMPRESS_URL had a protocol,
            # remove it and the hostname from our path.
            local_path = local_path.replace(self.protocol + self.host, "", 1)
        # remove url fragment, if any
        local_path = local_path.rsplit("#", 1)[0]
        # remove querystring, if any
        local_path = local_path.rsplit("?", 1)[0]
        # Now, we just need to check if we can find
        # the path from COMPRESS_URL in our url
        if local_path.startswith(self.url_path):
            local_path = local_path.replace(self.url_path, "", 1)
        # Re-build the local full path by adding root
        filename = os.path.join(self.root, local_path.lstrip("/"))
        return os.path.exists(filename) and filename

    def add_suffix(self, url):
        filename = self.guess_filename(url)
        if not filename:
            return url
        if settings.COMPRESS_CSS_HASHING_METHOD is None:
            return url
        if not url.startswith(SCHEMES):
            return url

        suffix = None
        if settings.COMPRESS_CSS_HASHING_METHOD == "mtime":
            suffix = get_hashed_mtime(filename)
        elif settings.COMPRESS_CSS_HASHING_METHOD in ("hash", "content"):
            suffix = get_hashed_content(filename)
        else:
            raise FilterError(
                "COMPRESS_CSS_HASHING_METHOD is configured "
                "with an unknown method (%s)." % settings.COMPRESS_CSS_HASHING_METHOD
            )
        fragment = None
        if "#" in url:
            url, fragment = url.rsplit("#", 1)
        if "?" in url:
            url = "%s&%s" % (url, suffix)
        else:
            url = "%s?%s" % (url, suffix)
        if fragment is not None:
            url = "%s#%s" % (url, fragment)
        return url

    def _converter(self, url):
        if url.startswith(("#", "data:")):
            return url
        elif url.startswith(SCHEMES):
            return self.add_suffix(url)
        full_url = posixpath.normpath("/".join([str(self.directory_name), url]))
        if self.has_scheme:
            full_url = "%s%s" % (self.protocol, full_url)
        full_url = self.add_suffix(full_url)
        return self.post_process_url(full_url)

    def post_process_url(self, url):
        """
        Extra URL processing, to be overridden in subclasses.
        """
        return url

    def url_converter(self, matchobj):
        quote = matchobj.group(1)
        converted_url = self._converter(matchobj.group(2))
        return "url(%s%s%s)" % (quote, converted_url, quote)

    def src_converter(self, matchobj):
        quote = matchobj.group(1)
        converted_url = self._converter(matchobj.group(2))
        return "src=%s%s%s" % (quote, converted_url, quote)


class CssRelativeFilter(CssAbsoluteFilter):
    """
    Do similar to ``CssAbsoluteFilter`` URL processing
    but add a *relative URL prefix* instead of ``settings.COMPRESS_URL``.
    """

    run_with_compression_disabled = True

    def post_process_url(self, url):
        """
        Replace ``settings.COMPRESS_URL`` URL prefix with  '../' * (N + 1)
        where N is the *depth* of ``settings.COMPRESS_OUTPUT_DIR`` folder.

        E.g. by default ``settings.COMPRESS_OUTPUT_DIR == 'CACHE'``,
        the depth is 1, and the prefix will be '../../'.

        If ``settings.COMPRESS_OUTPUT_DIR == 'my/compiled/data'``,
        the depth is 3, and the prefix will be '../../../../'.

        Example:

        - original file URL: '/static/my-app/style.css'
        - it has an image link: ``url(images/logo.svg)``
        - compiled file URL: '/static/CACHE/css/output.abcdef123456.css'
        - replaced image link URL: ``url(../../my-app/images/logo.svg)``
        """
        old_prefix = self.url
        if self.has_scheme:
            old_prefix = "{}{}".format(self.protocol, old_prefix)
        # One level up from 'css' / 'js' folder
        new_prefix = ".."
        # N levels up from ``settings.COMPRESS_OUTPUT_DIR``
        new_prefix += "/.." * len(
            list(
                filter(
                    None, os.path.normpath(settings.COMPRESS_OUTPUT_DIR).split(os.sep)
                )
            )
        )
        return re.sub("^{}".format(old_prefix), new_prefix, url)



import os
import brotli

from django.core.files.base import ContentFile
from django.core.files.storage import storages
from django.core.files.storage.base import Storage
from django.test import TestCase
from django.test.utils import override_settings
from django.utils.functional import LazyObject

from compressor import storage
from compressor.conf import settings
from compressor.css import CssCompressor
from compressor.tests.test_base import css_tag
from compressor.tests.test_templatetags import render


class GzipStorage(LazyObject):
    def _setup(self):
        self._wrapped = storages.create_storage({
            "BACKEND": "compressor.storage.GzipCompressorFileStorage"
        })


class BrotliStorage(LazyObject):
    def _setup(self):
        self._wrapped = storages.create_storage({
            "BACKEND": "compressor.storage.BrotliCompressorFileStorage"
        })


class DummyPathNotImplementedStorage(Storage):
    """
     A dummy storage backend that mimics a remote storage that does not implement
     `.path()` e.g. `storages.backends.s3.S3Storage`.
    """
    def exists(self, name):
        return True

    def path(self, name):
        raise NotImplementedError


@override_settings(COMPRESS_ENABLED=True)
class StorageTestCase(TestCase):
    def setUp(self):
        self.default_storage = storage.default_storage
        storage.default_storage = GzipStorage()
        storage.brotli_storage = BrotliStorage()

    def tearDown(self):
        storage.default_storage = self.default_storage

    def test_gzip_storage(self):
        storage.default_storage.save("test.txt", ContentFile("yeah yeah"))
        self.assertTrue(
            os.path.exists(os.path.join(settings.COMPRESS_ROOT, "test.txt"))
        )
        self.assertTrue(
            os.path.exists(os.path.join(settings.COMPRESS_ROOT, "test.txt.gz"))
        )

    def test_brotli_storage(self):
        payload = ",".join([str(i) for i in range(1000)]).encode()
        chunk_size = 1024
        storage.brotli_storage.save("test.txt", ContentFile(payload))
        self.assertTrue(
            os.path.exists(os.path.join(settings.COMPRESS_ROOT, "test.txt"))
        )
        self.assertTrue(
            os.path.exists(os.path.join(settings.COMPRESS_ROOT, "test.txt.br"))
        )
        decompressed_data = b""
        br_decompressor = brotli.Decompressor()
        with open(os.path.join(settings.COMPRESS_ROOT, "test.txt.br"), "rb") as f:
            for data in iter(lambda: f.read(chunk_size), b""):
                decompressed_data += br_decompressor.process(data)
        self.assertEqual(payload, decompressed_data)

    def test_css_tag_with_storage(self):
        template = """{% load compress %}{% compress css %}
        <link rel="stylesheet" href="{{ STATIC_URL }}css/one.css" type="text/css">
        <style type="text/css">p { border:5px solid white;}</style>
        <link rel="stylesheet" href="{{ STATIC_URL }}css/two.css" type="text/css">
        {% endcompress %}
        """
        context = {"STATIC_URL": settings.COMPRESS_URL}
        out = css_tag("/static/CACHE/css/output.e701f86c6430.css")
        self.assertEqual(out, render(template, context))

    def test_duplicate_save_overwrites_same_file(self):
        filename1 = self.default_storage.save("test.txt", ContentFile("yeah yeah"))
        filename2 = self.default_storage.save("test.txt", ContentFile("yeah yeah"))
        self.assertEqual(filename1, filename2)
        self.assertNotIn("_", filename2)

    def test_offline_manifest_storage(self):
        storage.default_offline_manifest_storage.save(
            "test.txt", ContentFile("yeah yeah")
        )
        self.assertTrue(
            os.path.exists(os.path.join(settings.COMPRESS_ROOT, "CACHE", "test.txt"))
        )
        # Check that the file is stored at the same default location as before
        # the new manifest storage.
        self.assertTrue(self.default_storage.exists(os.path.join("CACHE", "test.txt")))


class CompressorFileNameTestCase(TestCase):
    @override_settings(
        COMPRESS_ENABLED=True,
        DEBUG=False,
        COMPRESS_STORAGE=(
            "compressor.tests.test_storages.DummyPathNotImplementedStorage"
        ),
    )
    def test_storage_without_path_fallback(self):
        """
        Remote storages not implementing path need a fallback to a private
        instance of CompressorFileStorage. This must not be dependent on
        project settings.
        """
        old_default_storage = storage.default_storage
        storage.default_storage = storage.DefaultStorage()

        css = (
            '<link rel="stylesheet" href="/static/css/one.css" type="text/css" />'
        )
        compressor = CssCompressor("css", css)
        try:
            # Remote storage would raise NotImplementedError if fallback is
            # unsuccessful.
            compressor.get_filename("css/one.css")
        finally:
            storage.default_storage = old_default_storage
